#!/usr/bin/env python
#
# processing_functions.py - Processing functions
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains definitions of processing functions - functions which
may be specifeid in the processing table.


A processing function may perform any sort of processing on one or more
variables. A processing function may add, remove, or manipulate the columns of
the :class:`DataTable`.


All processing functions must accept the following as their first two
positional arguments:


 - The :class:`.DataTable` object, containing references to the data, variable,
   and processing table.
 - A list of integer ID of the variables to process.


Furthermore, all processing functions must return one of the following:

 - ``None``, indicating that no columns are to be added or removed.

 - A ``list`` (must be a ``list``) of :class:`.Column` objects describing the
   columns that should be removed from the data.

 - A ``tuple`` (must be a ``tuple``) of length 2, containing:

    - A list of ``pandas.Series`` that should be added to the data.

    - A list of variable IDs to use for each new ``Series``. This list must
      have the same length as the list of new ``Series``, but if they are not
      associated with any specific variable, ``None`` may be used.

 - A ``tuple`` of length 3, containing:

    - List of columns to be removed
    - List of ``Series`` to be added
    - List of variable IDs for each new ``Series``.

 - A ``tuple`` of length 4, containing the above, and:

    - List of dicts associated with each of the new ``Series``. These will be
      passed as keyword arguments to the :class:`.Column` objects that
      represent each of the new ``Series``.

The following processing functions are defined:

 .. autosummary::
   :nosignatures:

   removeIfSparse
   removeIfRedundant
   binariseCategorical
   expandCompound
   createDiagnosisColumns
   removeField
"""


import functools       as ft
import itertools       as it
import                    logging
import                    collections

import numpy            as np
import pandas           as pd
import pandas.api.types as pdtypes

from typing import List, Optional, Any

from . import processing_functions_core as core
from . import                              util
from . import                              custom
from . import                              datatable
from . import                              loadtables


log = logging.getLogger(__name__)


@custom.processor()
def removeIfSparse(
        dtable     : datatable.DataTable,
        vids       : List[int],
        minpres    : Optional[float] = None,
        minstd     : Optional[float] = None,
        mincat     : Optional[float] = None,
        maxcat     : Optional[float] = None,
        abspres    : bool            = True,
        abscat     : bool            = True,
        naval      : Optional[Any]   = None,
        ignoreType : bool            = False
) -> List[datatable.Column]:
    """removeIfSparse([minpres][, minstd][, mincat][, maxcat][, abspres][, abscat][, naval])
    Removes columns deemed to be sparse.

    Removes columns for all specified variables if they fail a sparsity test.
    The test is based on the following criteria:

     - The number/proportion of non-NA values must be greater than
       or equal to ``minpres``.

     - The standard deviation of the data must be greater than ``minstd``.

     - For integer and categorical types, the number/proportion of the largest
       category must be greater than ``mincat``.

     - For integer and categorical types, the number/proportion of the largest
       category must be less than ``maxcat``.

    If **any** of these criteria are **not met**, the data is considered to be
    sparse. Each criteria can be disabled by passing in ``None`` for the
    relevant parameter.

    The ``minstd`` test is only performed on numeric columns, and the
    ``mincat``/``maxcat`` tests are only performed on integer/categorical
    columns.

    If ``abspres=True`` (the default), ``minpres`` is interpreted as an
    absolute count. If ``abspress=False``, ``minpres`` is interpreted as a
    proportion.

    Similarly, If ``abscat=True`` (the default), ``mincat`` and ``maxcat`` are
    interpreted as absolute counts. Otherwise ``mincat`` and ``maxcat`` are
    interpreted as proportions

    The ``naval`` argument can be used to customise the value to consider as
    "missing" - it defaults to ``np.nan``.
    """  # noqa

    # :arg ignoreType: Defaults to ``False``. If ``True``, all specified tests
    #                  are run regardless of the types of the ``vids``. Only
    #                  used for testing.


    cols   = []
    series = []
    vtypes = []

    for vid in vids:

        if ignoreType: vtype = None
        else:          vtype = dtable.vartable.loc[vid, 'Type']

        vcols = dtable.columns(vid)

        cols  .extend(vcols)
        series.extend([dtable[:, c.name] for c in vcols])
        vtypes.extend([vtype] * len(vcols))

    log.debug('Checking %u columns for sparsity %s ...', len(series), vids[:5])

    func = ft.partial(core.isSparse,
                      minpres=minpres,
                      minstd=minstd,
                      mincat=mincat,
                      maxcat=maxcat,
                      abspres=abspres,
                      abscat=abscat,
                      naval=naval)

    with dtable.pool() as pool:
        results = pool.starmap(func, zip(series, vtypes))

    remove = []

    for col, (isSparse, reason, value) in zip(cols, results):
        if isSparse:
            log.debug('Dropping sparse column %s (%s: %f)',
                      col.name, reason, value)
            remove.append(col)

    if len(remove) > 0:
        log.debug('Dropping %u sparse columns: %s ...',
                  len(remove), [r.name for r in remove[:5]])

    return remove


@custom.processor()
def removeIfRedundant(dtable       : datatable.DataTable,
                      vids         : List[int],
                      corrthres    : float,
                      nathres      : float = None,
                      skipUnknowns : bool  = False,
                      precision    : str   = None,
                      pairwise     : bool  = False):
    """removeIfRedundant(corrthres, [nathres])
    Removes columns deemed to be redundant.

    Removes columns from the specified group of variables if they are found to
    be redundant with any other columns in the group.

    Redundancy is determined by calculating the correlation between all pairs
    of columns - columns with an absolute correlation greater than
    ``corrthres`` are identified as redundant.

    The test can optionally take the patterns oof missing values into account
    - if ``nathres`` is provided, the missingness correlation is also
    calculated between all column pairs. Columns must have absolute
    correlation greater than ``corrthres`` **and** absolute missingness
    correlation greater than ``nathres`` to be identified as redundant.

    The ``skipUnknowns`` option defaults to ``False``.  If it is set to
    ``True``, columns which are deemed to be redundant with respect to an
    unknown or uncategorised column are **not** dropped.

    The ``precision`` option can be set to either ``'double'`` (the default)
    or ``'single'`` - this controls whether 32 bit (single) or 64 bit (double)
    precision floating point is used for the correlation calculation. Double
    precision is recommended, as the correlation calculation algorithm can
    be unstable for data with large values (>10e5).
    """
    # :arg pairwise: Use alternative pairwise implementation. If ``pairwise``
    #                is ``True``, an alternative implementation is used which
    #                may be faster on data sets with high missingness
    #                correlation.


    # Ignore non-numeric columns
    cols     = list(it.chain(*[dtable.columns(v) for v in vids]))
    cols     = [c for c in cols if pdtypes.is_numeric_dtype(dtable[:, c.name])]
    colnames = [c.name for c in cols]
    data     = dtable[:, colnames]

    with np.errstate(divide='ignore'):
        if pairwise:
            redundant = _pairwiseRemoveIfRedundant(
                dtable, data, corrthres, nathres)
        else:
            redundant = _removeIfRedundant(
                dtable, data, corrthres, nathres, precision)

    redundant = util.dedup(sorted(redundant))

    if skipUnknowns:
        copy = []
        for idxa, idxb in redundant:
            colb = cols[idxa]
            bvid = colb.vid
            cats = loadtables.variableCategories(dtable.cattable, [bvid])[bvid]

            if 'unknown' in cats or 'uncategorised' in cats:
                namea = colnames[idxa]
                nameb = colnames[idxb]
                log.debug('Column %s is redundant with %s, but %s is '
                          'unknown / uncategorised, so %s will not be '
                          'dropped', namea, nameb, nameb, namea)
            else:
                copy.append((idxa, idxb))
        redundant = copy

    if len(redundant) > 0:
        log.debug('Dropping %u redundant columns', len(redundant))

    return [cols[r[0]] for r in redundant]


def _removeIfRedundant(dtable, data, corrthres, nathres=None, precision=None):
    """Default fast implementation of redundancy check. Used when the
    ``pairwise`` option to :func:`removeIfRedundant` is ``False``.

    :arg dtable:    The :class:`.DataTable` containing all data
    :arg data:      ``pandas.DataFrame`` containing the data to check
    :arg corrthres: Correlation threshold - see :func:`.redundantColumns`.
    :arg nathres:   Missingness correlation threshold - see
                    :func:`.redundantColumns`.
    :arg precision: Floating point precision -``'single'`` or ``'double'``.
    :returns:       Sequence of tuples of column indices, where each tuple
                    ``(a, b)`` indicates that column ``a`` is redundant with
                    respect to column ``b``.
    """

    return core.matrixRedundantColumns(data, corrthres, nathres, precision)


def _pairwiseRemoveIfRedundant(dtable, data, corrthres, nathres=None):
    """Alternative implementation of redundancy check. Used when the
    ``pairwise`` option to :func:`removeIfRedundant` is ``True``.

    :arg dtable:    The :class:`.DataTable` containing all data
    :arg data:      ``pandas.DataFrame`` containing the data to check
    :arg corrthres: Correlation threshold - see :func:`.redundantColumns`.
    :arg nathres:   Missingness correlation threshold - see
                    :func:`.redundantColumns`.
    :returns:       Sequence of tuples of column indices, where each tuple
                    ``(a, b)`` indicates that column ``a`` is redundant with
                    respect to column ``b``.
    """

    ncols = len(data.columns)

    # If we are correlating missingness,
    # we use the naCorrelation function
    # to identify all of the column pairs
    # which are na-correlated - the pairs
    # which fail this test do not need to
    # be subjected to the correlation test
    # (and therefore pass the redundancy
    # check)
    if nathres is not None:
        nacorr = core.naCorrelation(pd.isna(data), nathres)

        # As the matrix is symmetric, we can
        # drop column pairs where x >= y.
        nacorr   = np.triu(nacorr, k=1)
        colpairs = np.where(nacorr)
        colpairs = np.vstack(colpairs).T

    # Otherwise we generate an array
    # containing indices of all column
    # pairs.
    else:
        xs, ys   = np.triu_indices(ncols, k=1)
        colpairs = np.vstack((xs, ys)).T

    # we need at least
    # one pair of columns
    if len(colpairs) == 0:
        return []

    # evaluate all pairs at once
    if not dtable.parallel:
        log.debug('Checking %u columns for redundancy', ncols)
        redundant = core.pairwiseRedundantColumns(
            data, colpairs, corrthres=corrthres)

    # evaluate in parallel
    else:
        # Split the column pairs
        # into njobs chunks, and
        # run in parallel
        chunksize  = int(np.ceil(len(colpairs) / dtable.njobs))
        pairchunks = [colpairs[i:i + chunksize]
                      for i in range(0, len(colpairs), chunksize)]

        log.debug('Checking %u columns for redundancy (%u tasks)',
                  ncols, len(pairchunks))

        with dtable.pool() as pool:
            results = []
            for i, chunk in enumerate(pairchunks):

                # We can pass the full dataframe
                # to each task, as it should be
                # read-accessible via shared memory.
                token  = 'task {} / {}'.format(i + 1, len(pairchunks))
                result = pool.apply_async(
                    core.pairwiseRedundantColumns,
                    kwds=dict(data=data,
                              colpairs=chunk,
                              corrthres=corrthres,
                              token=token))
                results.append(result)

            # wait for the tasks to complete,
            # and gather the results (indices
            # of redundant columns) into a list
            redundant = []
            for result in results:
                redundant.extend(result.get())

    return redundant


# auxvids tells the processing runner the
# "take" argument refers to other variables
# which are not processed, but are needed
# to perform the processing.
#
# "filterMissing" tells the processing
# runner *not* to remove variables which
# are not present in the data from the list
# of vids that are passed in - we do our
# own check here.
@custom.processor(auxvids=['take'], filterMissing=False)
def binariseCategorical(dtable,
                        vids,
                        acrossVisits=False,
                        acrossInstances=True,
                        minpres=None,
                        nameFormat=None,
                        replace=True,
                        take=None,
                        fillval=None,
                        replaceTake=True):
    """binariseCategorical([acrossVisits][, acrossInstances][, minpres][, nameFormat][, replace][, take][, fillval][, replaceTake])
    Replace a categorical column with one binary column per category.

    Binarises categorical variables - replaces their columns with one new
    column for each unique value, containing ``1`` for subjects with that
    value, and ``0`` otherwise. Thos procedure is applied independently to all
    variables that are specified.

    The ``acrossVisits`` option controls whether the binarisation is applied
    across visits for each variable. It defaults to ``False``, meaning that
    the binarisation is applied separately to the columns within each
    visit. If set to ``True``, the binarisation will be applied to the columns
    for all visits. Similarly, the ``acrossInstances`` option controls whether
    the binarisation is applied across instances. This defaults to ``True``,
    which is usually desirable - for example, data field `41202
    <https://biobank.ctsu.ox.ac.uk/crystal/field.cgi?id=41202>`_ contains
    multiple ICD10 diagnoses, separated across different instances.

    If the ``minpres`` option is specified, it is used as a threshold -
    categorical values with less than this many occurrences will not be added
    as columns.

    The ``nameFormat`` argument controls how the new data columns should be
    named - it must be a format string using named replacement fields
    ``'vid'``, ``'visit'``, ``'instance'``, and ``'value'``. The ``'visit'``
    and ``'instance'`` fields may or may not be necessary, depending on the
    value of the ``acrossVisits`` and ``acrossInstances`` arguments.

    The default value for the ``nameFormat`` string is as follows:

    ================ =================== ======================================
    ``acrossVisits`` ``acrossInstances`` ``nameFormat``
    ================ =================== ======================================
    ``False``        ``False``           ``'{vid}-{visit}.{instance}_{value}'``
    ``False``        ``True``            ``'{vid}-{visit}.{value}'``
    ``True``         ``False``           ``'{vid}-{value}.{instance}'``
    ``True``         ``True``            ``'{vid}-{value}'``
    ================ =================== ======================================

    The ``replace`` option controls whether the original un-binarised columns
    should be removed - it defaults to ``True``, which will cause them to be
    removed.

    By default, the new binary columns (one for each unique value in the
    input columns) will contain a ``1`` indicating that the value is present,
    or a ``0`` indicating its absence. As an alternative to this, the ``take``
    option can be used to specify *another* variable from which to take values
    when populating the output columns. ``take`` may be set to a variable ID,
    or sequence of variable IDs (one for each of the input variables) to take
    values from. If provided, the generated columns will have values from the
    column(s) of this variable, instead of containinng binary 0/1 values.

    A ``take`` variable must have columns that match the columns of the
    corresponding variable (by both visits and instances).

    If ``take`` is being used, the ``fillval`` option can be used to specify the
    the value to use for ``False`` / ``0`` rows. It defaults to ``np.nan``.

    The ``replaceTake`` option is similar to ``replace`` - it controls whether
    the columns associated with the ``take`` variables are removed (``True`` -
    the defailt), or retained.
    """  # noqa

    # get groups of columns for vid, grouped
    # according to acrossVisits/acrossInstances
    def gatherColumnGroups(vid):
        colgroups = []
        visits    = dtable.visits(   vid)
        instances = dtable.instances(vid)
        if not (acrossVisits or acrossInstances):
            for visit, instance in it.product(visits, instances):
                colgroups.append(dtable.columns(vid, visit, instance))
        elif acrossInstances and (not acrossVisits):
            for visit in visits:
                colgroups.append(dtable.columns(vid, visit))
        elif (not acrossInstances) and acrossVisits:
            for instance in instances:
                colgroups.append(dtable.columns(vid, instance=instance))
        else:
            colgroups = [dtable.columns(vid)]
        return colgroups

    defaultNameFormat = {
        (False, False) : '{vid}-{visit}.{instance}_{value}',
        (False, True)  : '{vid}-{visit}.{value}',
        (True,  False) : '{vid}-{value}.{instance}',
        (True,  True)  : '{vid}-{value}',
    }

    if nameFormat is None:
        nameFormat = defaultNameFormat[acrossVisits, acrossInstances]

    # if take is a single vid or None,
    # we turn it into [take] * len(vids)
    if not isinstance(take, collections.abc.Sequence):
        take = [take] * len(vids)

    if len(take) != len(vids):
        raise ValueError('take must be either None, a single variable ID, '
                         'or a list of variable IDs, one for each of the '
                         'main vids.')

    remove     = []
    newseries  = []
    newvids    = []
    newcolargs = []

    for vid, takevid in zip(vids, take):

        if (not dtable.present(vid)) or \
           (takevid is not None and not dtable.present(takevid)):
            log.warning('Variable %u (or take: %s) is not present in the '
                        'data set - skipping the binariseCategorical step',
                        vid, takevid)
            continue

        colgrps = gatherColumnGroups(vid)

        if takevid is None: takegrps = [None] * len(colgrps)
        else:               takegrps = gatherColumnGroups(takevid)

        for cols, takecols in zip(colgrps, takegrps):

            log.debug('Calling binariseCategorical (vid: %i, '
                      '%u columns)', vid, len(cols))

            if takecols is None: tkdata = None
            else:                tkdata = dtable[:, [c.name for c in takecols]]

            data              = dtable[:, [c.name for c in cols]]
            binarised, values = core.binariseCategorical(data,
                                                         minpres=minpres,
                                                         take=tkdata,
                                                         token=vid,
                                                         njobs=dtable.njobs)

            if replace:                                remove.extend(cols)
            if replaceTake and (takecols is not None): remove.extend(takecols)

            for col, val in zip(binarised.T, values):

                fmtargs = {
                    'vid'      : str(int(cols[0].vid)),
                    'visit'    : str(int(cols[0].visit)),
                    'instance' : str(int(cols[0].instance)),
                    'value'    : str(val)
                }

                series = pd.Series(
                    col,
                    index=dtable.index,
                    name=nameFormat.format(**fmtargs))

                # The value is stored on each Column object as
                # an attribute "binariseCategorical_value".
                # This may be used by other processing
                # functions (see e.g. createDiagnosisColumns
                # with the binarised=True option).  The value
                # is also stored as "metadata", which is used
                # by the --description_file cli option - see
                # funpack.main.generateDescription).
                colargs = {
                    'metadata'                  : val,
                    'binariseCategorical_value' : val,
                    'basevid'                   : takevid,
                    'fillval'                   : fillval
                }

                newvids   .append(vid)
                newcolargs.append(colargs)
                newseries .append(series)

    return remove, newseries, newvids, newcolargs


@custom.processor()
def expandCompound(dtable, vids, nameFormat=None, replace=True):
    """expandCompound([nameFormat][, replace])
    Expand a compound column into a set of columns, one for each value.

    Expands compound variables into a set of columns, one for each value.
    Rows with different number of values are padded with ``np.nan``.

    This procedure is applied independently to each column of each specified
    variable.

    The ``nameFormat`` option can be used to control how the new columns
    should be named - it must be a format string using named replacement
    fields ``'vid'``, ``'visit'``, ``'instance'``, and ``'index'``.  The
    default value for ``nameFormat`` is ``'{vid}-{visit}.{instance}_{index}'``.

    The ``replace`` option controls whether the original columns are removed
    (``True`` - the default), or retained.
    """

    if nameFormat is None:
        nameFormat = '{vid}-{visit}.{instance}_{index}'

    columns   = list(it.chain(*[dtable.columns(v) for v in vids]))
    newseries = []
    newvids   = []

    for column in columns:

        data    = dtable[:, column.name]
        newdata = core.expandCompound(data)

        for i in range(newdata.shape[1]):

            coldata = newdata[:, i]
            name    = nameFormat.format(vid=column.vid,
                                        visit=column.visit,
                                        instance=column.instance,
                                        index=i)

            newvids  .append(column.vid)
            newseries.append(pd.Series(coldata,
                                       index=dtable.index,
                                       name=name))

    if replace: return columns, newseries, newvids
    else:       return          newseries, newvids


@custom.processor(auxvids=["primvid", "secvid"])
def createDiagnosisColumns(dtable,
                           vids,
                           primvid,
                           secvid,
                           replace=True,
                           primfmt=None,
                           secfmt=None,
                           binarised=False):
    """createDiagnosisColumns(primvid, secvid)
    Create binary columns for (e.g.) ICD10 codes, denoting them as either primary or secondary.

    This function is intended for use with data fields containing ICD9, ICD10,
    OPSC3, and OPSC4 diagnosis codes.  The UK Biobank publishes these
    diagnosis/operative procedure codes twice:

      - The codes are published in a "main" data field containing all codes
      - The codes are published again in two other data fields, containing
        separate "primary" and "secondary" codes.

    =====  ===============  ==================  ====================
    Code   Main data field  Primary data field  Secondary data field
    =====  ===============  ==================  ====================
    ICD10  41270            41202               41204
    ICD9   41271            41203               41205
    OPSC4  41272            41200               41210
    OPSC3  41273            41256               41258
    =====  ===============  ==================  ====================

    For example, this function may be applied to the ICD10 diagnosis codes
    like so::

        41270 createDiagnosisColumns(41202, 41204)

    When applied to one of the main data fields, (e.g. 41270 - ICD10
    diagnoses), this function will create two new columns for every unique
    ICD10 diagnosis code:

     - the first column contains a 1 if the code corresponds to a primary
       diagnosis (i.e. is also in 41202).
     - the second column contains a 1 if the code corresponds to a secondary
       diagnosis (i.e. is also in 41204).

    The ``replace`` option defaults to ``True`` - this causes the primary and
    secondary code columns to be removed from the data set.

    The ``binarised`` option defaults to ``False``, which causes this function
    to expect the input columns to be in their raw format, as described in the
    UKB showcase
    (e.g. https://biobank.ctsu.ox.ac.uk/crystal/field.cgi?id=41270). The
    ``binarised`` option can be set to ``True``, which will allow this
    function to be applied to data fields which have been passed through the
    :func:`binariseCategorical` function.
    """  # noqa

    if len(vids) == 0:
        return

    if len(vids) != 1:
        raise ValueError('createDiagnosisColumns can only be '
                         f'applied to one datafield [{vids}]')

    if primfmt is None: primfmt = '{primvid}-{code}.primary'
    if secfmt  is None: secfmt  = '{secvid}-{code}.secondary'

    # create separate data frames containing
    # only primary/secondary code columns
    mainvid  = vids[0]
    maincols = dtable.columns(mainvid)
    pricols  = dtable.columns(primvid)
    seccols  = dtable.columns(secvid)
    pridf    = dtable[:, [c.name for c in pricols]]
    secdf    = dtable[:, [c.name for c in seccols]]

    log.debug('Counting unique values across %u columns...', len(maincols))

    # Original data field has already been passed
    # through binariseCategorical - we have one
    # input column per unique code (and can get
    # those codes from the Column objects)
    if binarised:
        uniq = [c.binariseCategorical_value for c in maincols]
        uniq = sorted(pd.unique(uniq))
    else:
        uniq = [dtable[:, c.name].dropna() for c in maincols]
        uniq = sorted(pd.unique(pd.concat(uniq, ignore_index=True)))

    log.debug('Identifying primary/secondary codes [%s = %s + %s]...',
              mainvid, primvid, secvid)
    pribin = np.zeros((dtable.shape[0], len(uniq)), dtype=np.uint8)
    secbin = np.zeros((dtable.shape[0], len(uniq)), dtype=np.uint8)

    for i, code in enumerate(uniq):
        pribin[:, i] = (pridf == code).any(axis=1)
        secbin[:, i] = (secdf == code).any(axis=1)

    fmtargs    = {'mainvid' : mainvid,
                  'primvid' : primvid,
                  'secvid'  : secvid}
    pribincols = [primfmt.format(code=code, **fmtargs) for code in uniq]
    secbincols = [secfmt .format(code=code, **fmtargs) for code in uniq]
    pribincols = [pd.Series(d, index=dtable.index, name=c)
                  for d, c in zip(pribin.T, pribincols)]
    secbincols = [pd.Series(d, index=dtable.index, name=c)
                  for d, c in zip(secbin.T, secbincols)]

    add     = pribincols + secbincols
    addvids = [primvid] * len(uniq) + [secvid] * len(uniq)
    colargs = [{'metadata' : c} for c in uniq] * 2

    if replace: remove = pricols + seccols
    else:       remove = []

    return remove, add, addvids, colargs


@custom.processor()
def removeField(dtable, vids):
    """removeField()
    Remove all columns associated with one or more datafields/variables.

    This function can be used to simply remove columns from the data set. This
    can be useful if a variable is required for some processing step, but is
    not needed in the final output file.
    """
    return list(it.chain(*[dtable.columns(v) for v in vids]))



#!/usr/bin/env python

import numpy as np

from funpack import cleaner

# Cleaner functions are passed:
#
#   - dtable: An object which provides access to the data set.
#   - vid:    The variable ID of the column(s) to be cleaned.
#
# Cleaner functions should modify the data in-place.
@cleaner()
def drop_odd_values(dtable, vid):

    # Remember that a variable may be
    # associated with multiple columns
    cols = dtable.columns(vid)

    # the columns() method returns a list of
    # Column objects,  each of which contains
    # information about one column in the data.
    for col in cols:
        col = col.name
        mask = dtable[:, col] % 2 != 0
        dtable[mask, col] = np.nan



#!/usr/bin/env python
#
# cleaning_functions.py - Cleaning functions.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains definitions of cleaning functions - functions
which may be specified in the ``Clean`` column of the variable table,
and which are appiled during the data import stage.

The following cleaning functions are available:

.. autosummary::
   :nosignatures:

   remove
   keepVisits
   fillVisits
   fillMissing
   flattenHierarchical
   codeToNumeric
   parseSpirometryData
   makeNa


All cleaning functions (with two exceptions - :func:`remove` and
:func:`keepVisits`, explained below) will be passed the following as their
first positional arguments, followed by any arguments specified in the
variable table:

 - The :class:`.DataTable` object, containing references to the data, variable,
   and processing table.
 - The integer ID of the variable to process

These functions are expected to perform in-place pre-processing on the data.


The :func:`remove` and :func:`keepVisits` functions are called before the data
is loaded, as they are used to determine which columns should be loaded
in. They therefore have a different function signature to the other cleaning
functions, which are called after the data has been loaded. These functions
are passed the following positional arguments, followed by arguments specified
in the variable table:

 - The variable table, containing metadata about all known variables

 - The integer ID of the variable to be checked

 - A list of :class:`.Column` objects denoting all columns
   associated with this variable that are present in the data

These functions are expected to return a modified list containing the columns
that should be loaded.
"""


import functools as ft
import              logging

import numpy            as np
import pandas           as pd
import pandas.api.types as pdtypes

from . import util
from . import icd10
from . import custom
from . import hierarchy
from . import expression


log = logging.getLogger(__name__)


@custom.cleaner()
def remove(vartable, vid, columns):
    """remove()
    Remove (do not load) all columns associated with ``vid``.

    :arg vartable: Variable table
    :arg vid:      Integer variable ID.
    :arg columns:  List of all :class:`.Column` objects associated with
                   ``vid``.
    :returns:      An empty list.
    """
    return []


@custom.cleaner()
def keepVisits(vartable, vid, columns, *tokeep):
    """keepVisits(visit[, visit, ...])
    Only load columns for ``vid`` which correspond to the specified visits.

    This test is only applied to variables which have an *instancing* equal to
    2:

    https://biobank.ctsu.ox.ac.uk/crystal/instance.cgi?id=2

    Such variables are not associated with a specific visit, so it makes no
    sense to apply this test to them. See the
    :func:`.loadtables.addNewVariable` function for more information on
    variable instancing.

    :arg vartable: Variable table

    :arg vid:      Integer variable ID.

    :arg columns:  List of all :class:`.Column` objects that are associated
                   with ``vid``.

    :arg tokeep:   Visit IDs (integers), or ``'first'``, or ``'last'``,
                   indicating that the first or last visit should be loaded.

    :returns:      List of columns that should be loaded.
    """

    # variables which don't follow instancing
    # 2 do not have columns that correspond
    # to specific visits.
    instancing = vartable.loc[vid, 'Instancing']
    if instancing != 2:
        return columns

    keep      = []
    minVisit  = min([c.visit for c in columns])
    maxVisit  = max([c.visit for c in columns])

    for col in columns:

        test = [col.visit]

        if col.visit == minVisit: test.append('first')
        if col.visit == maxVisit: test.append('last')

        if col not in keep and any([v in tokeep for v in test]):
            keep.append(col)

    return keep


@custom.cleaner()
def fillVisits(dtable, vid, method='mode'):
    """fillVisits([method=(mode|mean)])
    Fill missing visits from available visits.

    For a variable with multiple visits, fills missing values from the visits
    that do contain data. The ``method`` argument can be set to one of:

      - ``'mode'`` (the default) - fill missing visits with the most
        frequently occurring value in the available visits
      - ``'mean'`` - fill missing visits with the mean of the values in the
        available visits
    """

    if method not in ('mode', 'mean'):
        raise ValueError('Unknown method: {}'.format(method))

    instances = dtable.instances(vid)

    for instance in instances:
        columns = [c.name for c in dtable.columns(vid, instance=instance)]
        view    = dtable[:, columns]

        if   method == 'mode': repvals = view.mode('columns')[0]
        elif method == 'mean': repvals = view.mean('columns')

        log.debug('Filling NA values in columns %s across visits', columns)

        for col in columns:
            dtable[:, col] = dtable[:, col].fillna(repvals)


@custom.cleaner()
def fillMissing(dtable, vid, value):
    """fillMissing(fill_value)
    Fill missing values with a constant.

    Fills all missing values in columns for the given variable with
    ``fill_value``.
    """
    columns = [c.name for c in dtable.columns(vid)]
    log.debug('Filling NA values in columns %s with %i', columns, value)
    for col in columns:
        dtable[:, col] = dtable[:, col].fillna(value)


@custom.cleaner()
def flattenHierarchical(dtable,
                        vid,
                        level=None,
                        name=None,
                        numeric=False,
                        convertNumeric=False):
    """flattenHierarchical([level][, name][, numeric][, convertNumeric])
    Replace leaf values with parent values in hierarchical variables.

    For hierarchical variables such as the ICD10 disease categorisations,
    this function replaces leaf values with a parent value.

    The ``level`` argument allows the depth of the parent value to be selected
    - ``0`` (the default) replaces a value with the top-level (i.e. its most
    distant) parent value, ``1`` the second-level parent value, etc.

    If, for a particular value, the ``level`` argument is greater than the
    number of ancestors of the value, the value is unchanged.

    By default, ``flattenHierarchical`` expects to be given coding values.  For
    example, for `data field 41202
    <https://biobank.ctsu.ox.ac.uk/crystal/field.cgi?id=41202>`_ (ICD10
    diagnoses), coding ``A009`` (Cholera, unspecified) would be replaced with
    ``Chapter I`` (Certain infectious and parasitic diseases).

    The ``numeric`` argument tells the function to expect numeric node IDs (as
    output by :func:`codeToNumeric`), rather than coding labels. This allows
    ``flattenHierarchical`` to be used in conjunction with ``codeToNumeric``,
    e.g.::

        -cl 41202 "codeToNumeric,flattenHierarchical(numeric=True)"

    Continuing the above example, the ``codeToNumeric`` function would replace
    ``A009`` with its node ID ``2890``, and
    ``flattenHierarchical(numeric=True)`` would replace ``2890`` with ``10``
    (the node ID for ``Chapter I``).

    The ``convertNumeric`` argument is a shortcut for the above scenario -
    using this implies ``numeric=True``, and will cause the ``codeToNumeric``
    routine to be called automatically. The following example is equivalent
    to the above::

        -cl 41202 "flattenHierarchical(convertNumeric=True)"

    The ``name`` option is unnecessary in most cases, but may be used to
    specify an encoding scheme to use. Currently, ``'icd9'``, ``'icd10'``,
    ``'opsc3'`` and ``'opsc4'`` are accepted.
    """

    if convertNumeric:
        numeric = True
        codeToNumeric(dtable, vid, name)

    if level is None:
        level = 0

    columns = [c.name for c in dtable.columns(vid)]
    vhier   = hierarchy.getHierarchyFilePath(dtable, vid, name)
    vhier   = hierarchy.loadHierarchyFile(vhier)

    # Hierarchy.parents returns a list of all
    # parents of a node - we want to return
    # the one at the requested level/depth
    def index_parent(parents, val):

        if pdtypes.is_number(parents) and pd.isna(parents):
            return val

        # the list returned by Hierarchy.parents
        # is ordered from closest to most distant
        # parent, so we index from the end.
        idx = len(parents) - 1 - level

        # If the specified level is out of
        # bounds, we return the original value
        if idx < 0: return val
        else:       return parents[idx]

    log.debug('Flattening hierarchical data in columns %s (level %i)',
              columns, level)

    if numeric: lookupFunc = vhier.parentIDs
    else:       lookupFunc = vhier.parents

    for col in columns:
        data           = dtable[:, col]
        parents        = data.map(lookupFunc, na_action='ignore')
        parents        = parents.combine(data, index_parent)
        dtable[:, col] = parents


@custom.cleaner()
def codeToNumeric(dtable, vid, name=None, scheme='node'):
    """codeToNumeric()
    Convert hierarchical coding labels to numeric equivalents (node IDs).

    Given a hierarchical variable which contains alpha-numeric codings, this
    function will replace the codings with numeric equivalents.

    For example, UKB data field 41270 (ICD10 diagnoses) uses `data coding 19
    <https://biobank.ctsu.ox.ac.uk/crystal/coding.cgi?id=19>`_, which uses
    ICD10 alpha-numeric codes to denote medical diagnoses.

    For data field 41270, ``codeToNumeric`` would replace coding ``D730``
    with the corresponding node ID ``22180`` when using the ``'node'``
    conversion scheme, or ``39730`` when using the ``'alpha'`` scheme.

    The ``name`` option is unnecessary in most cases, but may be used to
    specify which encoding scheme to use. Currently, ``'icd9'``, ``'icd10'``,
    ``'opsc3'`` and ``'opsc4'`` are accepted.

    The ``scheme`` option determines the conversion method - it can be set
    to one of:

      - ``'node'`` (default): Convert alpha-numeric codes to the corresponding
        numeric node IDs, as specified in the UK Biobank showcase.

      - ``'alpha'``: Convert alpha-numeric codes to integer numbers, using an
        immutable scheme that will not change over time - this scheme uses the
        :func:`.icd10.toNumeric` function.

    .. warning:: Note that node IDs (which are used by the ``'node'``
                 conversion scheme) are not present for non-leaf nodes in all
                 hierarchical UK Biobank data fields (in which case this
                 function will return `nan`). Also note that node IDS **may
                 change across UK Biobank showcase releases**.

    .. warning:: The ``'alpha'`` conversion scheme is intended for use with
                 ICD10 alpha-numeric codes (and similar coding schemes). This
                 scheme may result in unrepresentable integer values if used
                 with longer strings.
    """

    hier = hierarchy.getHierarchyFilePath(dtable, vid, name=name)
    hier = hierarchy.loadHierarchyFile(hier)

    if scheme == 'node':
        convert = ft.partial(hierarchy.codeToNumeric, hier=hier)
    elif scheme == 'alpha':
        convert = icd10.toNumeric
    else:
        raise ValueError(f'Invalid value for scheme option: {scheme}')

    cols  = [c.name for c in dtable.columns(vid)]
    codes = dtable[:, cols]

    log.debug('Converting hierarchy codes to numeric for '
              'variable %i [%u columns]', vid, len(cols))

    for col in cols:
        dtable[:, col] = codes[col].map(convert)

    if name is None:
        coding = dtable.vartable.loc[vid, 'DataCoding']
        ncmap  = {c : n for n, c in hierarchy.HIERARCHY_DATA_NAMES.items()}
        name   = ncmap.get(coding, None)

    if name == 'icd10':
        for col in cols:
            icd10.storeCodes(codes[col].tolist())


@custom.cleaner()
def parseSpirometryData(dtable, vid):
    """parseSpirometryData()
    Parse spirometry (lung volume measurement) data.

    Parses values which contain spirometry (lung volume measurement) test
    data.

    Columns for UK Biobank data fields `3066
    <https://biobank.ctsu.ox.ac.uk/crystal/field.cgi?id=3066>`_ and `10697
    <https://biobank.ctsu.ox.ac.uk/crystal/field.cgi?id=10697>`_ contain
    comma-separated time series data. This function converts the text values
    into ``numpy`` arrays, and potentially re-orders instances of the variable
    within each visit - subjects potentially underwent a series of tests, with
    each test stored as a separate instance within one visit. However, the
    instance order does not necessarily correspond to the test order.
    """
    visits = dtable.visits(vid)

    # applied to each row
    def fixOrder(vals):

        def parseBlow(val):
            # the first entry is "blowN",
            # where N is the blow number
            try:
                tkn = val[:10].split(',', maxsplit=1)[0]
                return int(tkn[4:])
            except Exception:
                return np.nan
        bnumvals = [(parseBlow(v), v) for v in vals]
        bnumvals = sorted(bnumvals)
        return pd.Series([bv[1] for bv in bnumvals],
                          index=vals.index)

    def parseValue(val):
        # first entry is "blowN", second
        # entry is number of samples.
        # Ignore them both, turn the
        # remainder into a numpy array
        try:
            val = val.split(',', maxsplit=2)[2]
            return np.fromstring(val, sep=',')
        except Exception:
            return np.nan

    for visit in visits:
        cols = dtable.columns(vid, visit=visit)
        cols = [c.name for c in cols]

        data = dtable[:, cols]
        data = data.apply(fixOrder,   axis=1)
        for col in cols:
            data.loc[:, col] = data.loc[:, col].apply(parseValue)

        dtable[:, cols] = data


@custom.cleaner()
def makeNa(dtable, vid, expr):
    """makeNa(expression)
    Replace values which pass the expression with NA.

    All values in the columns of the variable for which ``func`` evaluates to
    ``True`` are replaced with ``nan``.

    The ``expression`` must be of the form ``'<operator> <value>'``, where
    ``<operator>`` is one of ``==``, ``!=``, ``<``, ``<=``, ``>``, ``>=``, or
    ``contains`` (only applicable to non-numeric variables). For example,
    ``makeNa('> 25')`` would cause all values greater than 25 to be replaced
    with ``nan``.
    """

    expr = expression.Expression('v{} {}'.format(vid, expr))
    cols = dtable.columns(vid)

    for col in cols:
        col               = col.name
        mask              = expr.evaluate(dtable[:], {vid : col})
        dtable[mask, col] = np.nan



#!/usr/bin/env python
#
# __init__.py - funpack package
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


__version__ = '3.5.2'
"""The ``funpack`` versioning scheme roughly follows Semantic Versioning
conventions.
"""


from .util      import (findConfigDir,  # noqa
                        findConfigDirs)
from .custom    import (loader,         # noqa
                        sniffer,
                        formatter,
                        exporter,
                        processor,
                        metaproc,
                        cleaner)
from .datatable import (DataTable,      # noqa
                        Column)


def version():
    """Return the FUNPACK version string. If running from an installed
    copy, simply returns :attr:`__version__`. Otherwise, returns the
    result of calling ``git describe --dirty --tags``.
    """
    import               shlex
    import os.path    as op
    import subprocess as sp

    rootdir = op.join(op.dirname(__file__), '..')
    cmd     = shlex.split('git describe --dirty --tags')

    try:
        result  = sp.run(cmd, capture_output=True, check=True, cwd=rootdir)
        return result.stdout.decode().strip()
    except Exception:
        # assume that we're running from an install
        return __version__



#!/usr/bin/env python
#
# test_expression.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import io
import textwrap  as tw
import itertools as it

import pytest

import pyparsing as pp

import pandas as pd
import numpy as np

import funpack.expression as expression

from . import gen_DataTable, gen_DataTableFromDataFrame



def test_Expression():

    _test_data = tw.dedent("""
    index 10 20 30
    1 1 2 3
    2 4 5 6
    3 7  9
    4 10 11 12
    5 13  15
    """).strip().replace(' ', '\t')

    _test_cols = {
        10 : '10',
        20 : '20',
        30 : '30',
    }


    _expr_tests = [
        (      'v10 == 1',   [1, 0, 0, 0, 0]),
        (      'v10 != 1',   [0, 1, 1, 1, 1]),
        (      'v10 >  7',   [0, 0, 0, 1, 1]),
        (      'v10 >= 7',   [0, 0, 1, 1, 1]),
        (      'v10 <  7',   [1, 1, 0, 0, 0]),
        (      'v10 <= 7',   [1, 1, 1, 0, 0]),
        (    '~(v10 == 1)',  [0, 1, 1, 1, 1]),
        (    '~(v10 != 1)',  [1, 0, 0, 0, 0]),
        (  'all(v10 != 1)',  [0, 1, 1, 1, 1]),
        (  'any(v10 != 1)',  [0, 1, 1, 1, 1]),
        ('~(all(v10 != 1))', [1, 0, 0, 0, 0]),
        ('~(any(v10 != 1))', [1, 0, 0, 0, 0]),
        ('all(~(v10 != 1))', [1, 0, 0, 0, 0]),


        ('v20 == na', [0, 0, 1, 0, 1]),
        ('v20 != na', [1, 1, 0, 1, 0]),

        (  'v10 >= 7 && v30 < 10',                [0, 0, 1, 0, 0]),
        ('~(v10 >= 7 && v30 < 10)',               [1, 1, 0, 1, 1]),
        ( 'v10 >= 4 &&  v30 < 10  || v20 != na',  [1, 1, 1, 1, 0]),
        ('(v10 >= 4 &&  v30 < 10) || v20 != na',  [1, 1, 1, 1, 0]),
        ( 'v10 >= 4 && (v30 < 10  || v20 != na)', [0, 1, 1, 1, 0]),

        ('v10 > 7 || v20 > 7 || v30 < 7',             [1, 1, 0, 1, 1]),
        ('v10 > 4 && v20 > 5 && v30 < 14',            [0, 0, 0, 1, 0]),
        ('v10 > 9 || v10 < 3 || v20 > 7 || v30 > 10', [1, 0, 0, 1, 1]),

        # n.b. and has higher precedence than or
        ('v10 > 4 && v20 > 5 || v30 < 14',        [1, 1, 1, 1, 0]),

        # bad
        ('10 == 1',        'error'),
        ('10 ==',          'error'),
        ('v10',            'error'),
        ('v10 ==',         'error'),
        ('v10 1',          'error'),
        ('v10 == 1 &&',    'error'),
        ('v10 == 1 && 24', 'error'),
        ('abcde',          'error'),
    ]

    def vine(e):
        vs = []
        if 'v10' in e: vs.append(10)
        if 'v20' in e: vs.append(20)
        if 'v30' in e: vs.append(30)
        return vs

    data =  pd.read_csv(io.StringIO(_test_data), sep='\t')

    for expr, expected in _expr_tests:

        if expected == 'error':
            with pytest.raises(pp.ParseException):
                e = expression.Expression(expr)
            continue
        else:
            e = expression.Expression(expr)

        assert sorted(e.variables) == sorted(vine(expr))

        result = e.evaluate(data, _test_cols)

        assert len(result) == len(expected)
        assert all([bool(r) == bool(e) for r, e in zip(result, expected)])


def test_Expression_multiple_columns():
    data       = np.random.randint(1, 100, (20, 6))
    data[:, 0] = np.arange(1, 21)
    cols       = ['eid', '1-0.0', '1-1.0', '1-2.0', '2-0.0', '2-1.0']
    df         = pd.DataFrame({c : d for c, d in zip(cols, data.T)}).set_index('eid')

    # one column
    e = expression.Expression('v1 > 50')
    result = e.evaluate(df, {1 : '1-0.0'})
    assert np.all(result.flatten() == (data[:, 1] > 50))

    # multiple columns
    e = expression.Expression('v1 <= 75')
    result = e.evaluate(df, {1 : ['1-0.0', '1-1.0']})
    assert np.all(result == (data[:, 1:3] <= 75))

    e = expression.Expression('v1 != 23')
    result = e.evaluate(df, {1 : ['1-0.0', '1-1.0', '1-2.0']})
    assert np.all(result == (data[:, 1:4] != 23))

    # multi-column multi-var - number
    # of columns per var must match
    e = expression.Expression('v1 > 50 && v2 < 50')
    result = e.evaluate(df, {1 : ['1-0.0', '1-1.0'],
                                 2 : ['2-0.0', '2-1.0']})
    exp = (data[:, 1:3] > 50) & (data[:, 4:] < 50)
    assert np.all(result == exp)

    # use any/all to collapse across columns
    e = expression.Expression('all(v1 > 50 && v2 < 50)')
    result = e.evaluate(df, {1 : ['1-0.0', '1-1.0'],
                             2 : ['2-0.0', '2-1.0']})
    exp = ((data[:, 1:3] > 50) & (data[:, 4:] < 50)).all(axis=1)
    assert np.all(result == exp)

    e = expression.Expression('any(v1 > 50 && v2 < 50)')
    result = e.evaluate(df, {1 : ['1-0.0', '1-1.0'],
                             2 : ['2-0.0', '2-1.0']})
    exp = ((data[:, 1:3] > 50) & (data[:, 4:] < 50)).any(axis=1)
    assert np.all(result == exp)

    e = expression.Expression('any(v1 > 50) && all(v2 < 50)')
    result = e.evaluate(df, {1 : ['1-0.0', '1-1.0', '1-2.0'],
                             2 : ['2-0.0', '2-1.0']})
    exp = (data[:, 1:4] > 50).any(axis=1) & (data[:, 4:] < 50).all(axis=1)
    assert np.all(result == exp)

    # column length mismatch - each variable
    # will be combined with logical OR, and
    # then expression evaluated on the 1D vectors
    e = expression.Expression('v1 > 50 && v2 < 50')
    exp = (data[:, 1:4] > 50).any(axis=1) & (data[:, 4:] < 50).any(axis=1)
    result = e.evaluate(df, {1 : ['1-0.0', '1-1.0', '1-2.0'],
                             2 : ['2-0.0', '2-1.0']})
    assert np.all(result == exp)


def makexprs(exprstrs):
    exprs = []
    for exprstr in exprstrs:
        expr = [expression.Expression(e) for e in exprstr.split(',')]
        if len(expr) == 1:
            expr = expr[0]
        exprs.append(expr)
    return exprs


def test_calculateExpressionEvaluationOrder():

    vids = [1, 2, 3, 4, 5, 7]
    exprs = [
        'v3 == 0',             # 1 depends on 3
        'v3  > 2',             # 2 depends on 3
        'v4 != na',            # 3 depends on 4
        'v5 != na, v6 == na',  # 4 depends on 5 and 6
        'v6 < 0',              # 5 depends on 6
        'v8 == 34'             # 7 depends on 8
    ]

    expected = [
        (4, [1, 2]),
        (3, [3]),
        (2, [4]),
        (1, [5, 7]),
    ]

    exprs = makexprs(exprs)
    result = expression.calculateExpressionEvaluationOrder(vids, exprs)
    assert result == expected


    # 1 depends on 2 and 3
    # 2 depends on 3
    vids = [1, 2]
    exprs = ['v2 == 20, v3 > 400',
             'v3 < 200']
    expected = [(2, [1]), (1, [2])]
    exprs = makexprs(exprs)
    result = expression.calculateExpressionEvaluationOrder(vids, exprs)
    assert result == expected

    # 1 depends on 2 and 3
    # 2 depends on 3
    vids = [1, 2]
    exprs = ['v3 > 400, v2 == 20',
             'v3 < 200']
    expected = [(2, [1]), (1, [2])]
    exprs = makexprs(exprs)
    result = expression.calculateExpressionEvaluationOrder(vids, exprs)
    assert result == expected


def test_calculateExpressionEvaluationOrder_circularDependency():
    # error on self-dependency
    with pytest.raises(ValueError):
        vids  = [1, 2]
        exprs = ['v1 == 1', 'v4 == 42']
        expression.calculateExpressionEvaluationOrder(vids, makexprs(exprs))

    # error on circular-dependency
    with pytest.raises(ValueError):
        vids  = [1, 2]
        exprs = ['v2 == 1', 'v1 == 2']
        expression.calculateExpressionEvaluationOrder(vids, makexprs(exprs))

    # length mismatch
    with pytest.raises(ValueError):
        vids  = [1, 2, 3]
        exprs = ['v2 == 1']
        expression.calculateExpressionEvaluationOrder(vids, makexprs(exprs))


# fsl/funpack#25
def test_calculateExpressionEvaluationOrder_unordered_vids():

    #    4
    #   / \
    #  3   \
    #  |\   \
    #  2 \   |
    #  \  |  |
    #   \ | /
    #    \|/
    #     1
    vids  = [1, 2, 3]
    exprs = [
        'v2 == 0 || v3 != 0 && v4 == 0', # 1 depends on 2, 3, and 4
        'v3 == 0 && v4 != 0',            # 2 depends on 3 and 4
        'v4 == 0',                       # 3 depends on 4
    ]

    expected = [
        (3, [1]),
        (2, [2]),
        (1, [3]),
    ]

    exprs = makexprs(exprs)

    for idxs in it.permutations(range(3)):
        ivids  = [vids[ i] for i in idxs]
        iexprs = [exprs[i] for i in idxs]
        result = expression.calculateExpressionEvaluationOrder(ivids, iexprs)
        assert result == expected


def test_Expresssion_non_numeric():
    test_data = tw.dedent("""
    index,10,20,30
    1,a,abc,a123
    2,b,def,a 101
    3,c,,b252
    4,d,jkl,b745
    5,e,,b254
    """).strip()

    test_cols = {
        10 : '10',
        20 : '20',
        30 : '30',
    }

    expr_tests = [
        ('v10 ==       "a"',     [1, 0, 0, 0, 0]),
        ("v10 ==       'a'",     [1, 0, 0, 0, 0]),
        ("v20 !=       'jkl'",   [1, 1, 1, 0, 1]),
        ('v20 ==       na',      [0, 0, 1, 0, 1]),
        ('v20 !=       na',      [1, 1, 0, 1, 0]),
        ('v30 ==       "a 101"', [0, 1, 0, 0, 0]),
        ('v30 contains "b2"',    [0, 0, 1, 0, 1]),
        ('v30 contains "b22"',   [0, 0, 0, 0, 0]),


        # bad
        ('v10 >  "a"', 'error'),
        ('v10 <= "a"', 'error'),
        ('v10 == a',  'error'),
        ('v10 == "a', 'error'),
        ('v10 == a"', 'error'),
    ]


    data =  pd.read_csv(io.StringIO(test_data), sep=',')

    for expr, expected in expr_tests:

        if expected == 'error':
            with pytest.raises(pp.ParseException):
                e = expression.Expression(expr)
            continue
        else:
            e = expression.Expression(expr)

        result = e.evaluate(data, test_cols)

        assert len(result) == len(expected)
        assert all([bool(r) == bool(e) for r, e in zip(result, expected)])


def test_Expresssion_datetime():
    testdata = tw.dedent("""
    eid,10,20
    1,2020-01-01,2020-01-01 01:00:00
    2,2020-02-01,2020-01-01 02:00:00
    3,2020-03-01,2020-01-01 03:00:00
    4,2020-04-01,2020-01-01 04:00:00
    """)

    test_cols = {
        10 : '10',
        20 : '20',
    }

    expr_tests = [
        ('v10 <= 2020-03-01',          [1, 1, 1, 0]),
        ('v10 == 2020-01-01',          [1, 0, 0, 0]),
        ('v10 >  2020-02-01',          [0, 0, 1, 1]),
        ('v20 == 2020-01-01 02:00:00', [0, 1, 0, 0]),
        ('v20 <  2020-01-01 03:00:00', [1, 1, 0, 0]),
        ('v20 >= 2020-01-01 02:00:00', [0, 1, 1, 1]),
    ]

    data =  pd.read_csv(io.StringIO(testdata),
                        sep=',',
                        parse_dates=['10', '20']).set_index('eid')

    for expr, expected in expr_tests:

        if expected == 'error':
            with pytest.raises(pp.ParseException):
                e = expression.Expression(expr)
            continue
        else:
            e = expression.Expression(expr)

        result = e.evaluate(data, test_cols)

        assert len(result) == len(expected)
        assert all([bool(r) == bool(e) for r, e in zip(result, expected)])



#!/usr/bin/env python
#
# This script refreshes the UKB showcase schema files that are built into
# FUNPACK. I run it by hand when releasing a new version of FUNPACK.
#


import itertools  as it
import subprocess as sp
import os.path    as op
import               glob
import               shlex


def download_file(url, dest):
    print(f'{url} -> {dest}')
    cmd = f'wget -O {dest} {url}'
    sp.run(shlex.split(cmd), stdout=sp.DEVNULL, stderr=sp.DEVNULL)


def main():
    basedir = op.join(op.dirname(__file__), '..', '..')
    datadir = op.join(basedir, 'funpack', 'data')

    baseurl = 'biobank.ctsu.ox.ac.uk/'

    download_file(f'{baseurl}/ukb/scdown.cgi?fmt=txt&id=1',
                  op.join(datadir, 'field.txt'))
    download_file(f'{baseurl}/ukb/scdown.cgi?fmt=txt&id=2',
                  op.join(datadir, 'encoding.txt'))

    codings = it.chain(
        glob.glob(op.join(datadir, 'coding',    '*.tsv')),
        glob.glob(op.join(datadir, 'hierarchy', '*.tsv')))

    for coding in codings:
        # all files are called "coding<ID>.tsv"
        cid = op.basename(coding)[6:-4]
        url = f'{baseurl}/crystal/codown.cgi?id={cid}'
        download_file(url, coding)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# test_util.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import os
import os.path as op
import multiprocessing as mp
from unittest import mock
import contextlib as ctxlib

import pytest

import funpack.util as util

from . import tempdir, patch_logging, mktree


def test_findConfigFile():
    # cfgdirs are (in order of precedence):
    #   FUNPACK_CONFIG_DIR
    #   site.getsitepackages()
    #   dirname(funpack.util.__file__)
    env = os.environ.copy()
    env.pop('FUNPACK_CONFIG_DIR', None)
    with tempdir() as td, mock.patch.dict('os.environ', env, clear=True):

        cfgdir  = op.join(td, 'cfgdir')
        sitedir = op.join(td, 'sitepack')
        fpdir   = op.join(td, 'fpdir')

        cfgdircfg  = op.join(cfgdir,                        'myconfig.cfg')
        sitedircfg = op.join(sitedir, 'funpack', 'configs', 'myconfig.cfg')
        fpdircfg   = op.join(fpdir,              'configs', 'myconfig.cfg')
        cfgdirplg  = op.join(cfgdir,                        'myplugin.py')
        sitedirplg = op.join(sitedir, 'funpack', 'plugins', 'myplugin.py')
        fpdirplg   = op.join(fpdir,              'plugins', 'myplugin.py')

        mktree((cfgdircfg, sitedircfg, fpdircfg,
                cfgdirplg, sitedirplg, fpdirplg))

        @ctxlib.contextmanager
        def mock_cfgdir():
            with mock.patch.dict(os.environ, FUNPACK_CONFIG_DIR=cfgdir):
                yield
        @ctxlib.contextmanager
        def mock_sitepackages():
            with mock.patch('site.getsitepackages', return_value=[sitedir]):
                yield
        @ctxlib.contextmanager
        def mock_funpack_dir():
            utilfile = op.join(fpdir, 'util.py')
            with mock.patch('funpack.util.__file__', utilfile):
                yield

        with mock_funpack_dir():
            assert util.findConfigFile('myconfig.cfg') == fpdircfg
            assert util.findConfigFile('myconfig')     == fpdircfg
            assert util.findPluginFile('myplugin.py')  == fpdirplg
            assert util.findPluginFile('myplugin')     == fpdirplg
        with mock_funpack_dir(), mock_sitepackages():
            assert util.findConfigFile('myconfig.cfg') == sitedircfg
            assert util.findConfigFile('myconfig')     == sitedircfg
            assert util.findPluginFile('myplugin.py')  == sitedirplg
            assert util.findPluginFile('myplugin')     == sitedirplg
        with mock_funpack_dir(), mock_sitepackages(), mock_cfgdir():
            assert util.findConfigFile('myconfig.cfg') == cfgdircfg
            assert util.findConfigFile('myconfig')     == cfgdircfg
            assert util.findPluginFile('myplugin.py')  == cfgdirplg
            assert util.findPluginFile('myplugin')     == cfgdirplg


def test_parseMatlabRange():
    tests = [
        ('1',      [1]),
        ('1:10',   list(range(1, 11))),
        ('1:2:10', [1, 3, 5, 7, 9]),
        ('10:-1:1', list(range(10, 0, -1))),
        ('10:-2:0', [10, 8, 6, 4, 2, 0]),
    ]

    for rng, exp in tests:
        assert util.parseMatlabRange(rng) == exp

    with pytest.raises(ValueError): util.parseMatlabRange('1:')
    with pytest.raises(ValueError): util.parseMatlabRange('1:1:')
    with pytest.raises(ValueError): util.parseMatlabRange('1:1:1:')
    with pytest.raises(ValueError): util.parseMatlabRange('1:1:1:1')
    with pytest.raises(ValueError): util.parseMatlabRange('abcde')
    with pytest.raises(ValueError): util.parseMatlabRange('a:b')
    with pytest.raises(ValueError): util.parseMatlabRange('a:b:c')


def test_parseColumnName():

    tests = [

        ('0-0.0',       ( 0,   0,  0)),
        ('1-1.1',       ( 1,   1,  1)),
        ('10-10.10',    (10,  10, 10)),
        ('1--1.1',      ( 1,  -1,  1)),
        ('10--10.10',   (10, -10, 10)),
        ('1-1.ABC',     ( 1,   1, 'ABC')),
        ('1-1.ABC.E',   ( 1,   1, 'ABC.E')),
        ('1-1.ABC E',   ( 1,   1, 'ABC E')),
        ('f.1.1.1',     ( 1,   1,  1)),
        ('f.10.10.10',  (10,  10, 10)),
        ('f.1..1.1',    ( 1,  -1,  1)),
        ('f.10..10.10', (10, -10, 10)),
        ('1.1',         ( 1,   0,  1)),
        ('1.ABC',       ( 1,   0, 'ABC')),
        ('1.ABC.E',     ( 1,   0, 'ABC.E')),
        ('1.ABC E',     ( 1,   0, 'ABC E')),
    ]

    for col, exp in tests:
        assert util.parseColumnName(col) == exp

    with pytest.raises(ValueError): util.parseColumnName('eid')
    with pytest.raises(ValueError): util.parseColumnName('f.eid')
    with pytest.raises(ValueError): util.parseColumnName('abc')
    with pytest.raises(ValueError): util.parseColumnName('10')
    with pytest.raises(ValueError): util.parseColumnName('10-')
    with pytest.raises(ValueError): util.parseColumnName('10-1')
    with pytest.raises(ValueError): util.parseColumnName('10-1-2')
    with pytest.raises(ValueError): util.parseColumnName('f.a')
    with pytest.raises(ValueError): util.parseColumnName('f.a.b')
    with pytest.raises(ValueError): util.parseColumnName('f.a.b.c')
    with pytest.raises(ValueError): util.parseColumnName('f.10.')
    with pytest.raises(ValueError): util.parseColumnName('f.10.10.')
    with pytest.raises(ValueError): util.parseColumnName('f.10.10.10.')
    with pytest.raises(ValueError): util.parseColumnName('f.10.10.10.10')
    with pytest.raises(ValueError): util.parseColumnName('f..10')
    with pytest.raises(ValueError): util.parseColumnName('f.10.10..10')


def test_generateColumnName():

    tests = [

        (( 0,   0,  0), '0-0.0'),
        (( 1,   1,  1), '1-1.1'),
        ((10,  10, 10), '10-10.10'),
        (( 1,  -1,  1), '1--1.1'),
        ((10, -10, 10), '10--10.10'),
    ]

    for var, exp in tests:
        assert util.generateColumnName(*var) == exp


@patch_logging
def test_timed():
    with util.timed('abc'):
        pass
    with util.timed('abc'):
        pass
    with util.timed('abc', fmt='abc %s %s'):
        pass
    with util.timed(fmt='abc %s %s'):
        pass


@patch_logging
def test_wc():
    text = '1\n2\n3\n4\n5\n'

    with tempdir():
        with open('file.txt', 'wt') as f:
            f.write(text)

        assert util.wc('file.txt') == 5

        with mock.patch('shutil.which', return_value=None):
            assert util.wc('file.txt') == 5


@patch_logging
def test_cat():

    f1  = '1\n2\n3\n'
    f2  = '4\n5\n6\n'
    f3  = '7\n8\n9\n'
    exp = f1 + f2 + f3

    with tempdir():
        with open('f1.txt', 'wt') as f: f.write(f1)
        with open('f2.txt', 'wt') as f: f.write(f2)
        with open('f3.txt', 'wt') as f: f.write(f3)

        util.cat(['f1.txt', 'f2.txt', 'f3.txt'], 'out1.txt')

        with mock.patch('shutil.which', return_value=None):
            util.cat(['f1.txt', 'f2.txt', 'f3.txt'], 'out2.txt')

        assert open('out1.txt').read() == exp
        assert open('out2.txt').read() == exp


def _checkproc(i):
    return util.inMainProcess()


def test_inMainProcess():
    pool = mp.Pool(8)
    assert util.inMainProcess()
    assert not any(pool.map(_checkproc, range(16)))



#!/usr/bin/env python
#
# processing_functions_core.py - Functions used by processing_functions.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains utility algorithms and functions used by the
:mod:`processing_functions` module.

 .. autosummary::
   :nosignatures:

   isSparse
   naCorrelation
   pairwiseRedundantColumns
   matrixRedundantColumns
   binariseCategorical
   expandCompound
"""


import                          enum
import                          logging
import                          collections
import functools             as ft
import multiprocessing       as mp
import multiprocessing.dummy as mpd

from typing import Optional, Tuple, List, Union, Any

import numpy                 as np
import pandas                as pd
import pandas.api.types      as pdtypes

import funpack.util          as util


log = logging.getLogger(__name__)


def isSparse(
        data    : pd.Series,
        ctype   : Optional[enum.Enum] = None,
        minpres : Optional[float]     = None,
        minstd  : Optional[float]     = None,
        mincat  : Optional[float]     = None,
        maxcat  : Optional[float]     = None,
        abspres : bool                = True,
        abscat  : bool                = True,
        naval   : Optional[Any]       = None
) -> Tuple[bool, Union[str, None], Any]:
    """Returns ``True`` if the given data looks sparse, ``False`` otherwise.

    Used by :func:`.removeIfSparse` - see that function for details.

    By default, the ``minstd`` test is only performed on numeric columns, and
    the ``mincat``/``maxcat`` tests are only performed on integer/categorical
    columns. This behaviour can be overridden by passing in a ``ctype`` of
    ``None``, in which case all speified tests will be performed.

    :arg data:     ``pandas.Series`` containing the data

    :arg ctype:    The series column type (one of the :attr:`.util.CTYPES`
                   values), or ``None`` to disable type-specific logic.

    :returns:      A tuple containing:

                    - ``True`` if the data is sparse, ``False`` otherwise.

                    - If the data is sparse, one of ``'minpres'``,
                      ``'minstd'``, ``mincat``, or ``'maxcat'``, indicating
                      the cause of its sparsity. ``None`` if the data is not
                      sparse.

                    - If the data is sparse, the value of the criteria which
                      caused the data to fail the test.  ``None`` if the data
                      is not sparse.
    """

    if naval is None: presmask = data.notnull()
    else:             presmask = data != naval

    present  = data[presmask]
    ntotal   = len(data)
    npresent = len(present)

    def fixabs(val, isabs, limit=ntotal, repl=npresent):

        # Turn proportion into
        # an absolute count
        if not isabs:
            val = val * limit
            if (val % 1) >= 0.5: val = np.ceil(val)
            else:                val = np.floor(val)

        # ignore the threshold if it is
        # bigger than the total data length
        if limit < val: return repl
        else:           return val

    iscategorical = ctype in (None,
                              util.CTYPES.integer,
                              util.CTYPES.categorical_single,
                              util.CTYPES.categorical_single_non_numeric,
                              util.CTYPES.categorical_multiple,
                              util.CTYPES.categorical_multiple_non_numeric)
    isnumeric     = ctype in (None,
                              util.CTYPES.continuous,
                              util.CTYPES.integer,
                              util.CTYPES.categorical_single,
                              util.CTYPES.categorical_multiple) and \
                    pdtypes.is_numeric_dtype(data)

    # not enough values
    if minpres is not None:
        if npresent < fixabs(minpres, abspres):
            return True, 'minpres', npresent

    # stddev is not large enough (for
    # numerical/categorical types)
    if isnumeric and minstd is not None:
        std = (present - present.mean()).std()
        if std <= minstd:
            return True, 'minstd', std

    # for categorical types
    if iscategorical and ((maxcat is not None) or (mincat is not None)):

        if maxcat is not None:
            maxcat = fixabs(maxcat, abscat, npresent, npresent + 1)
        if mincat is not None:
            mincat = fixabs(mincat, abscat, npresent)

        # mincat - smallest category is too small
        # maxcat - one category is too dominant
        uniqvals   = pd.unique(present)
        uniqcounts = [sum(present == u) for u in uniqvals]

        # Column is empty - force the test to fail
        if len(uniqcounts) == 0:
            if mincat is not None:
                mincat  = 1
                nmincat = 0
            if maxcat is not None:
                maxcat  = 0
                nmaxcat = 0
        else:
            nmincat = min(uniqcounts)
            nmaxcat = max(uniqcounts)

        if mincat is not None:
            if nmincat < mincat:
                return True, 'mincat', nmincat

        if maxcat is not None:
            if nmaxcat >= maxcat:
                return True, 'maxcat', nmaxcat

    return False, None, None


def naCorrelation(
        namask   : np.ndarray,
        nathres  : float,
        nacounts : Optional[np.ndarray] = None
) -> np.ndarray:
    """Compares the missingness correlation of every pair of columns in
    ``namask``.

    :arg namask:   2D ``numpy`` array of shape ``(nsamples, ncolumns)``
                   containing binary missingness classification

    :arg nathres:  Threshold above which column pairs should be classed
                   as correlated.

    :arg nacounts: 1D ``numpy`` array containing the umber of missing values
                   in each column. Calculated if not provided.

    :returns:      A ``(ncolumns, ncolumns)`` boolean ``numpy`` array
                   containing ``True`` for column pairs which exceed
                   ``nathres``, ``False`` otherwise.
    """

    log.debug('Calculating missingness correlation '
              '(%u columns)', namask.shape[1])

    if nacounts is None:
        nacounts = namask.sum(axis=0)

    # must be float32 to take
    # advantage of parallelism
    namask = namask.astype(np.float32)
    nacorr = np.corrcoef(namask.T)
    nacorr = np.abs(nacorr) > nathres

    # columns with no/all missing values
    # will contain invalid correlation
    # values. But we pass them because
    # these results will be combined with
    # a data correlation test.
    flatmask            = (nacounts == 0) | (nacounts == namask.shape[0])
    nacorr[flatmask, :] = True
    nacorr[:, flatmask] = True

    return nacorr


def pairwiseRedundantColumns(
        data      : pd.DataFrame,
        colpairs  : np.ndarray,
        corrthres : float,
        token     : Optional[str] = None
) -> List[Tuple[int, int]]:
    """Identifies redundant columns based on their correlation with each
    other by comparing each pair of columns one by one.

    :arg data:      ``pandas.DataFrame`` containing the data

    :arg colpairs:  ``numpy`` array of shape ``(N, 2)``, containing indices
                    of the column pairs to be evaluated.

    :arg corrthres: Correlation threshold - columns with a correlation greater
                    than this are identified as redundant.

    :arg token:     Identifier string for log messages.

    :returns:       Sequence of tuples of column indices, where each tuple
                    ``(a, b)`` indicates that column ``a`` is redundant with
                    respect to column ``b``.
    """

    if len(colpairs) == 0:
        return []

    if token is None: token = ''
    else:             token = '[{}] '.format(token)

    redundant = {}
    nacounts  = data.isna().sum(axis=0).to_numpy()

    # calculate correlation between column pairs
    for i, (coli, colj) in enumerate(colpairs):

        if i % 1000 == 0:
            log.debug('%sTesting column pair %u / %u',
                      token, i + 1, len(colpairs))

        datai = data.iloc[:, coli]
        dataj = data.iloc[:, colj]
        corr  = np.abs(datai.corr(dataj))

        # i and j are highly correlated -
        # flag the one with more missing
        # values as redundant
        if corr > corrthres:
            if nacounts[coli] > nacounts[colj]: drop, keep = coli, colj
            else:                               drop, keep = colj, coli

            if drop not in redundant:
                log.debug('%sColumn %s is redundant (correlation with %s: %f)',
                          token, data.columns[drop], data.columns[keep], corr)
                redundant[drop] = keep

    return list(redundant.items())


def matrixRedundantColumns(
        data      : pd.DataFrame,
        corrthres : float,
        nathres   : float = None,
        precision : str   = None) -> List[Tuple[int, int]]:
    """Identifies redundant columns based on their correlation with each
    other using dot products to calculate a correlation matrix.

    :arg data:      ``pandas.DataFrame`` containing the data

    :arg corrthres: Correlation threshold - columns with a correlation greater
                    than this are identified as redundant.

    :arg nathres:   Correlation threshold to use for missing values. If
                    provided, columns must have a correlation greater than
                    ``corrthres`` *and* a missing-value correlation greater
                    than ``nathres`` to be identified as redundant.

    :arg precision: ``'double'`` (the default) or ``'single'``, specifying
                    the floating point precision to use. Note that the
                    algorithm used to calculate the correlation values is
                    unstable for data with a range larger than ~10e5, so
                    double precision is used as default.

    :returns:       Sequence of tuples of column indices, where each tuple
                    ``(a, b)`` indicates that column ``a`` is redundant with
                    respect to column ``b``.
    """

    if len(data.columns) < 2:
        return []

    if precision == 'single':
        dtype = np.float32
    elif precision in (None, 'double'):
        dtype = np.float64
    else:
        raise ValueError(f'Invalid precision: {precision}')

    # Keep a copy of the column names for logging.
    # Create a 2D matrix containing all data
    columns = data.columns
    data    = data.to_numpy(dtype=dtype, copy=True)

    namask   = np.isnan(data)
    nacounts = namask.sum(axis=0)

    # missingness correlation
    if nathres is None: nacorr = None
    else:               nacorr = naCorrelation(namask, nathres, nacounts)

    log.debug('Calculating pairwise correlations '
              '(matrix shape: %s)', data.shape)

    # zero out nan elements
    data[namask] = 0

    # p=present elements
    namask  = (~namask).astype(dtype)
    Ap = Bp = namask
    A  = B  = data

    # sum x_i y_i
    xy = A.T @ B

    # number of items defined both in x and y
    n  = Ap.T @ Bp

    # mean values in x, calculated across items defined both in x and y
    # mean values in y, calculated across items defined both in x and y
    mx = (A.T @ Bp) / n
    my = (Ap.T @ B) / n

    # sum x^2_i, calculated across items defined both in x and y
    # sum y^2_i, calculated across items defined both in x and y
    x2 = (A * A).T @ Bp
    y2 = Ap.T @ (B * B)

    # sx, sy - standard deviations
    # sx = sqrt(x2 - n .* (mx.^2));
    # sy = sqrt(y2 - n .* (my.^2));
    sx         = x2 - n * (mx ** 2)
    sx[sx < 0] = 0
    sx         = np.sqrt(sx)

    sy         = y2 - n * (my ** 2)
    sy[sy < 0] = 0
    sy         = np.sqrt(sy)

    # correlation coefficient
    coef = (xy - n * mx * my) / (sx * sy)

    # ignore nans/infs, binarise. Keep a copy
    # of the raw correlations for logging.
    rawcoef                  = coef
    coef[~np.isfinite(coef)] = 0
    coef                     = np.abs(coef) > corrthres

    # columns must have a correlation greater than
    # corrthres and a missing-value correlation greater
    # than nathres to be identified as redundant.
    if nacorr is not None:
        coef = coef & nacorr

    # generate indices of correlated column
    # pairs; we only need to consider one half
    # of the triangle
    coef     = np.triu(coef, k=1)
    colpairs = np.vstack(np.where(coef))

    # for each correlated pair, we flag the
    # one with more missing values as redundant
    def correlatedPairs(coli, colj):
        if nacounts[coli] > nacounts[colj]: drop, keep = coli, colj
        else:                               drop, keep = colj, coli

        # coef is only the upper triangle
        if coli <= colj: corr = rawcoef[coli, colj]
        else:            corr = rawcoef[colj, coli]

        log.debug('Column %s is redundant (correlation with %s: %0.6f)',
                  columns[drop], columns[keep], corr)

        return drop, keep

    # Generate a sequence of pairs, where the first
    # element is the column to drop, and the second
    # element is the column it is redundant w.r.t.
    correlatedPairs = np.vectorize(correlatedPairs, [np.uint32, np.uint32])
    colsa, colsb    = correlatedPairs(colpairs[0], colpairs[1])

    if len(colsa) == 0:
        return []

    # Return only one pair for each column
    # that is to be dropped (the first
    # pair in the natural column ordering).
    idxs      = np.unique(colsa, return_index=True)[1]
    colsa     = colsa[idxs]
    colsb     = colsb[idxs]

    return list(zip(colsa, colsb))


def binariseCategorical(
        data    : pd.DataFrame,
        minpres : Optional[int]          = None,
        take    : Optional[pd.DataFrame] = None,
        token   : Optional[str]          = None,
        njobs   : Optional[int]          = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Takes one or more columns containing categorical data,, and generates a
    new set of binary columns (as ``np.uint8``), one for each unique
    categorical value, containing ``1`` in rows where the value was present,
    ``0`` otherwise.

    :arg data:    A ``pandas.DataFrame`` containing the input columns

    :arg minpres: Optional threshold - values with less than this number of
                  occurrences will not be included in the output

    :arg take:    Optional ``pandas.DataFrame`` containing values to use in
                  the output. Instead of using binary ``0``/``1`` values,
                  rows where a unique value is present will be populated with
                  the corresponding value from ``take``, and rows where a
                  unique value is not present will be populated with
                  ``np.nan``. Must contain the same number of columns (and
                  rows) as ``data``.

    :arg token:   Unique token to identify this function call, to make it
                  re-entrant (in case multiple calls are made in a parallel
                  execution environment).

    :arg njobs:   Number of jobs to parallelise tasks with - the
                  :func:`generateBinaryColumns` function is called in parallel
                  for different blocks of unique values.

    :returns:     A tuple containing:

                   - A ``(nrows, nvalues)`` ``numpy`` array containing the
                     generated binary columns

                   - A 1D ``numpy`` array of length ``nvalues`` containing
                     the unique values that are encoded in the binary columns.
    """

    if njobs is None: njobs = 1
    if njobs < 1:     njobs = 1

    if njobs == 1: Pool = mpd.Pool
    else:          Pool = mp.Pool

    if (take is not None) and take.shape != data.shape:
        takes = take.shape
        datas = data.shape
        takec = take.columns.difference(data.columns)
        datac = data.columns.difference(take.columns)
        raise ValueError('take {} must have same shape as data {}. '
                         'Column difference: {} -- {}'.format(
                             takes, datas, datac, takec))

    if minpres is None:
        minpres = 0

    if token is None: token = ''
    else:             token = '[{}] '.format(token)

    log.debug('%sCounting unique values across %u '
              'columns...', token, len(data.columns))

    cols    = [data[c]      for c in data.columns]
    cols    = [c[c.notna()] for c in cols]
    coluniq = [np.unique(c, return_counts=True) for c in cols]
    uniq    = collections.defaultdict(lambda : 0)

    for coluniq, colcounts in coluniq:
        for val, count in zip(coluniq, colcounts):
            uniq[val] += count

    # drop values with low occurrence
    uniq = [v for v, c in uniq.items() if c >= minpres]
    uniq = list(sorted(uniq))

    if len(uniq) == 0:
        return np.zeros((len(data), 0)), np.zeros((0,))

    log.debug('%sCounted %u unique values [minpres: %u]',
              token, len(uniq), minpres)

    # Prepare inputs for parallelising the
    # binarise using binariseUniqueValues
    data = data.to_numpy()

    # Figure out the type and fill value
    # of the output array. if take is
    # provided, we assume that every
    # column in it has the same dtype
    if take is None:
        dtype = np.uint8
        fill  = 0
    else:
        dtype = take.dtypes[0]
        take  = take.to_numpy()

        if   np.issubdtype(dtype, np.integer):    fill = 0
        elif np.issubdtype(dtype, np.datetime64): fill = np.datetime64('NAT')
        else:                                     fill = np.nan

    # We parallelise binarisation across
    # blocks of unique values - each call
    # to generateBinaryColumns will binarise
    # one block. Only read access is needed
    # for the uniq, data, and take arrays,
    # so they are made available at the
    # module level (and thus accessible to
    # the worker processes in shared parent
    # process memory).
    #
    # Write access is needed for the bindata
    # array (where the result is written),
    # so it is shared as a mp.RawArray.
    #
    # If take is None, bindata is a uint8
    # array; otherwise it is a take.dtype
    # array.  If take has a dtype of
    # np.datetime64, we need a hack, as we
    # can't create a sharedmem array of that
    # type (it is not supported by
    # numpy.ctypeslib.as_ctypes_type). So
    # for datetime, we make bindata uint64,
    # and then cast it back to datetime64
    # afterwards.
    if np.issubdtype(dtype, np.datetime64):
        cdtype = np.uint64
    else:
        cdtype = dtype

    # Shared array to store binarised
    # columns. If not parallelising,
    # we use a regular numpy array
    binshape = (len(data), len(uniq))

    if njobs > 1:
        rawbindata = mp.RawArray(np.ctypeslib.as_ctypes_type(cdtype),
                                 int(np.prod(binshape)))
        bindata    = np.ctypeslib.as_array(rawbindata).reshape(binshape)
    else:
        bindata    = np.empty(binshape, dtype=dtype)

    # make the inputs module-level accessible
    # before creating worker processess. We
    # use a unique token for this invocation
    # in case this function has been called
    # multiple times
    if not hasattr(generateBinaryColumns, 'inputs'):
        generateBinaryColumns.inputs = {}
    generateBinaryColumns.inputs[token] = (uniq, data, bindata, take)

    # create an offset into the uniq
    # list for each worker process
    valsperjob = int(np.ceil(len(uniq) / njobs))
    offsets    = np.arange(0, len(uniq), valsperjob)
    func       = ft.partial(generateBinaryColumns,
                            nvals=valsperjob,
                            fill=fill,
                            token=token)

    try:
        with Pool(njobs) as pool:
            pool.map(func, offsets)
            pool.close()
            pool.join()
    finally:
        generateBinaryColumns.inputs.pop(token)

    # cast the result if necessary
    if dtype != cdtype:
        bindata = bindata.astype(dtype)

    return bindata, uniq


def generateBinaryColumns(offset, nvals, fill, token):
    """Called by :func:`binariseCategorical`. Generates binarised columns for
    a subset of unique values for a data set.

    The following arguments are passed internally from
    :func:`binariseCategorical` to this function:

     - Sequence of unique values
     - Numpy array containing the data
     - Numpy array to store the output
     - Numpy array to take values from

    :arg offset: Offset into the sequence of unique values
    :arg nvals:  Number of unique values to binarise
    :arg fill:   Default value
    :arg token:  Unique token used to retrieve the data for one invocation of
                 this function.
    """

    # get refs to the shared inputs,
    uniq, data, bindata, take = generateBinaryColumns.inputs[token]

    # are all values being binarised in one go?
    allvals = (offset == 0) and (len(uniq) == nvals)

    # don't overflow on the last
    # block of unique values
    if offset + nvals > len(uniq):
        nvals = len(uniq) - offset

    # rather than writing to bindata directly
    # (which could be a mem-mapped file,
    # therefore slow), we write to a smaller
    # in-memory array and then copy afterwards
    # (but not if this function call is
    # binarising all unique values)
    if allvals: binblock = bindata
    else:       binblock = np.empty((len(data), nvals), bindata.dtype)

    binblock[:] = fill

    for i in range(nvals):

        v = uniq[i + offset]

        if (i + offset) % 100 == 0:
            log.debug('%sUnique value %u / %u [%s] ...',
                      token, i + offset + 1, len(uniq), v)

        mask = data == v

        # if take is None, each column is 1
        #  where the subject had an entry
        #  equal to that value, 0 otherwise.
        if take is None:
            binblock[:, i] = mask.any(axis=1)

        # if take is provided, each column
        # contains the value from take where
        # the subject had an entry equal to
        # that value, or nan otherwise.
        # If a subject had multiple entries
        # equal to value, the first
        # corresponding entry from take is
        # used (via this first function)
        else:
            rowmask              = mask.any(axis=1)
            idxs                 = np.argmax(mask, axis=1)
            values               = take[np.arange(take.shape[0]), idxs]
            binblock[rowmask, i] = values[rowmask]

    # copy results to the output array
    if not allvals:
        bindata[:, offset:offset + nvals] = binblock


def expandCompound(data : pd.Series) -> np.ndarray:
    """Takes a ``pandas.Series`` containing sequence data (potentially of
    variable length), and returns a 2D ``numpy`` array containing the parsed
    data.

    The returned array has shape ``(X, Y)``, where ``X`` is the number of rows
    in the data, and ``Y`` is the maximum number of values for any of the
    entries.

    :arg data: ``pandas.Series`` containing the compound data.
    :returns:   2D ``numpy`` array containing expanded data.
    """

    nrows = len(data)
    lens  = data.apply(len).values
    ncols = max(lens)

    # Create a 2D array from
    # rows of different lengths
    #
    # https://stackoverflow.com/a/32043366
    #
    # 2D mask array of shape (nrows, max(lens))
    # which identifies positions to be filled
    mask          = np.arange(ncols) < np.atleast_2d(lens).T
    newdata       = np.full((nrows, ncols), np.nan, dtype=np.float32)
    newdata[mask] = np.concatenate(data.values)

    return newdata



#!/usr/bin/env python
#
# test_config.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import site
import os
import shlex
import os.path as op

from unittest import mock

import funpack.config as config
import funpack.custom as custom

from . import clear_plugins, tempdir, touch

def get_fmrib_config_dir():
    for sitedir in site.getsitepackages():
        fmribdir = op.join(sitedir, 'funpack', 'configs', 'fmrib')
        if op.exists(fmribdir):
            return fmribdir

    raise RuntimeError('Cannot locate FMRIB configuration directory')


# This function is essentially testing funpack.util.findConfigFile
@clear_plugins
def test_parseArgs_configFilePaths():

    custom.registerBuiltIns()

    with tempdir() as td:
        cfgdir = op.join(td, 'a')
        os.makedirs(op.join(cfgdir, 'b'))
        with mock.patch.dict(os.environ, FUNPACK_CONFIG_DIR=cfgdir):
            cfgfile = op.join(cfgdir, 'b', 'myconfig.cfg')
            varfile = op.join(cfgdir, 'b', 'myvars.tsv')
            plgfile = op.join(cfgdir, 'b', 'myplugin.py')
            touch(cfgfile)
            touch(varfile)
            touch(plgfile)

            # full path
            args = config.parseArgs(shlex.split(f'-cfg {cfgfile} out in'))[0]
            assert args.config_file == [cfgfile]

            # rel path
            relp = op.join('a', 'b', 'myconfig.cfg')
            args = config.parseArgs(shlex.split(f'-cfg {relp} out in'))[0]
            assert args.config_file == [cfgfile]

            # rel to cfgdir
            relp = op.join('b', 'myconfig.cfg')
            args = config.parseArgs(shlex.split(f'-cfg {relp} out in'))[0]
            assert args.config_file == [cfgfile]

            # dotted rel to cfgdir
            args = config.parseArgs(shlex.split('-cfg b.myconfig out in'))[0]
            assert args.config_file == [cfgfile]

            argv = ['-vf', varfile,
                    '-vf', op.join('a', 'b', 'myvars.tsv'),
                    '-vf', op.join('b', 'myvars.tsv'),
                    '-vf', 'b.myvars',
                    'output',  'input']

            args = config.parseArgs(argv)[0]
            assert args.variable_file == [varfile, varfile, varfile, varfile]

            argv = ['-p', plgfile,
                    '-p', op.join('a', 'b', 'myplugin.py'),
                    '-p', op.join('b', 'myplugin.py'),
                    '-p', 'b.myplugin',
                    'output',  'input']
            args = config.parseArgs(argv)[0]
            assert args.plugin_file == [plgfile, plgfile, plgfile, plgfile]

@clear_plugins
def test_num_jobs():

    custom.registerBuiltIns()

    with mock.patch('multiprocessing.cpu_count', return_value=99):
        assert config.parseArgs('-nj -1 out in'.split())[0].num_jobs == 99
        assert config.parseArgs('-nj -5 out in'.split())[0].num_jobs == 99

    assert config.parseArgs('-nj 0 out in'.split())[0].num_jobs == 1
    assert config.parseArgs('-nj 1 out in'.split())[0].num_jobs == 1
    assert config.parseArgs('-nj 5 out in'.split())[0].num_jobs == 5


@clear_plugins
def test_multiple_config_file():

    custom.registerBuiltIns()
    with tempdir():

        with open('one.cfg',   'wt') as f:
            f.write('variable\t1\n')
            f.write('variable_file\tvf\n')
            f.write('processing_file\tpf1\n')
        with open('two.cfg',   'wt') as f:
            f.write('variable\t2\n')
            f.write('datacoding_file\tdf\n')
            f.write('processing_file\tpf2\n')
        with open('three.cfg', 'wt') as f:
            f.write('variable\t3\n')
            f.write('processing_file\tpf3\n')
        argv = '-cfg one.cfg -cfg two.cfg -cfg three.cfg out in'.split()
        args = config.parseArgsWithConfigFile(argv)[0]
        assert args.variable        == [1, 2, 3]
        assert args.variable_file   == ['vf']
        assert args.datacoding_file == ['df']
        assert args.processing_file ==  'pf3'


@clear_plugins
def test_recursive_config_file():

    custom.registerBuiltIns()
    with tempdir():

        with open('one.cfg',   'wt') as f:
            f.write('variable\t1\n')
            f.write('variable_file\tvf\n')
            f.write('processing_file\tpf1\n')
            f.write('config_file\ttwo.cfg\n')
        with open('two.cfg',   'wt') as f:
            f.write('variable\t2\n')
            f.write('datacoding_file\tdf\n')
            f.write('processing_file\tpf2\n')
            f.write('config_file\tthree.cfg\n')
            f.write('config_file\tfour.cfg\n')
        with open('three.cfg', 'wt') as f:
            f.write('variable\t3\n')
            f.write('processing_file\tpf3\n')
        with open('four.cfg', 'wt') as f:
            f.write('variable\t4\n')
            f.write('category_file\tcf\n')
            f.write('processing_file\tpf4\n')
        argv = '-cfg one.cfg out in'.split()
        args = config.parseArgsWithConfigFile(argv)[0]
        assert args.variable        == [1, 2, 3, 4]
        assert args.variable_file   == ['vf']
        assert args.datacoding_file == ['df']
        assert args.processing_file ==  'pf4'
        assert args.category_file   ==  'cf'



#!/usr/bin/env python
#
# processing.py - Cleaning and processing parsing and functions.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functionality for parsing and running "processes",
processing steps which are applied to one or more variables, and which may
result in columns being removed from, or new columns being added to, the
data table. These processes are specified in the processing table.


The parsing and execution functionality in this module are also used to manage
cleaning functions, which are specified in the ``Clean`` column of the
variable table. Definitions of all available (pre-)processing functions are in
the :mod:`cleaning_functions` and :mod:`.processing_functions` modules.


The :func:`processData` function is also defined here - it executes the
processes defined in the processing table.


Special processing functions can be applied to a variable's data by adding
them to the ``Clean`` and ``Process``  of the variable or processing
table respectively.  Processing is specified as a comma-separated list of
process functions - for example::


    process1, process2(), process3('arg1', arg2=1234)


The :func:`parseProcesses` function parses such a line, and returns a list of
:class:`Process` objects which can be used to query the process name and
arguments, and to run each process.


The processing table
--------------------


Within the processing table, the above process specification is preceeded by a
variable definition, which states which variables the process should be
applied to, and how they should be applied - either independently, or
together. This information is used to parallelise the processing where
possible.


For each process, the processing table contains a "process variable type", a
list of vids, and the process itself.  The process variable type is one of
the following:

  - No process variable type, simply a comma-separated list of VIDS, The
    process is applied to the specified vids::

        1,2,3 processName

  - ``'independent'``: The process is applied independently to the specified
    vids::

        independent,1,2,3 processName

  - ``'all'``: The process is applied to all vids::

        all processName

  - ``'all_independent'``: The process is applied independently to all vids::

       all_independent processName

  - ``'all_except'``: The process is applied to all vids except the specified
    ones::

       all_except,1,2,3 processName

  - ``'all_independent_except'``: The process is applied independently to all
    vids except the specified ones::

       all_independent_except,1,2,3 processName


For example, the :func:`.processing_functions.binariseCategorical` function
applies its logic independently to each variable, so it makes sense to specify
that it should be applied independently to a set of variables::

    independent,1,2,3,4,5 binariseCategorical

However, the :func:`.removeIfRedundant` function works on a collection of
variables (probably all variables in the data set), and this must be specified
in the processing table.

    all,1,2,3,4,5 removeIfRedundant(0.99)


Broadcasting arguments
----------------------

.. warning:: Broadcasting arguments are deprecated, and will be removed in
             FUNPACK 4.0.0. The alternative to broadcasting is for processing
             functions to perform their own parallelisation.

When a process is applied independently to more than one variable, the input
arguments to the process may need to be different for each variable. This can
be accomplished by using a _broadcast_ argument - simply prefix the argument
name with ``'broadcast_'``, and then specify a list containing the argument
values for each variable. For example, the following specification will result
in the :func:`.processing_functions.binariseCategorical` process being applied
independently to variables 1, 2, and 3, with values taken from variables 4, 5,
and 6 respectively.

    independent,1,2,3 binariseCategorical(broadcast_take=[4, 5, 6])

Note that broadcast arguments are useful as a performance optimisation - the
above specification is functionally equivalent to::

    1 binariseCategorical(take=4)
    2 binariseCategorical(take=5)
    3 binariseCategorical(take=6)

however the latter example cannot take advantage of parallelism in a
multi-core environment.
"""


import functools       as ft
import itertools       as it
import os.path         as op
import                    os
import                    glob
import                    logging
import                    tempfile
import                    warnings
import                    collections
import collections.abc as abc

import pyparsing as pp
import pandas    as pd

from . import util
from . import custom


log = logging.getLogger(__name__)


def processData(dtable):
    """Applies all processing specified in the processing table to the data.

    :arg dtable: The :class:`DataTable` instance.
    """

    for i in dtable.proctable.index:

        procs, vids, parallel = retrieveProcess(dtable, i)
        allvids               = list(it.chain(*vids))

        if len(allvids) == 0:
            continue

        # Run each process sequentially -
        # each process may be parallelised
        # by the runProcess function
        for proc in procs.values():

            log.debug('Running process %s on %u variables %s ...',
                      proc.name, len(allvids), allvids[:5])

            fmt = '[{} {} ...] completed in %s (%+iMB)'.format(
                proc.name, allvids[:5])

            with util.timed(proc.name, log, logging.DEBUG, fmt=fmt), \
                 tempfile.TemporaryDirectory() as workDir:
                runProcess(proc, dtable, vids, workDir, parallel)


def retrieveProcess(dtable, procIdx):
    """Used by :func:`processData`. Retrieves the process at index ``procIdx``
    in the processing table, and generates the variable groups that the
    process should be applied to.

    :arg dtable:  The :class:`.DataTable`
    :arg procIdx: Index into the processing table
    :returns:     A tuple containing:

                    - A dict of ``{ name : Process }`` mappings containing
                      the :class:`.Process` objects to be executed.

                    - A list of lists, each list containing a group of
                      variable IDs that the process should be applied to

                    - ``True`` if the process can be applied in parallel
                      across the variable groups, ``False`` otherwise.
    """

    i         = procIdx
    ptable    = dtable.proctable
    all_vids  = dtable.variables
    all_vids  = [v for v in all_vids if v != 0]

    # For each process, the processing table
    # contains a "process variable type",
    # a list of vids, and the process itself.
    # The pvtype is one of:
    #   - vids:                   apply the process to the specified vids
    #   - independent:            apply the process independently to the
    #                             specified vids
    #   - all:                    apply the process to all vids
    #   - all_independent:        apply the process independently to all
    #                             vids
    #   - all_except:             apply the process to all vids except the
    #                             specified ones
    #   - all_independent_except: apply the process independently to all
    #                             vids except the specified ones
    pvtype, vids = ptable.loc[i, 'Variable']
    procs        = ptable.loc[i, 'Process']

    # Build a list of lists of vids, with
    # each vid list a group of variables
    # that is to be processed together.

    # apply independently to every variable
    if pvtype in ('all_independent', 'all_independent_except'):
        if pvtype.endswith('except'): exclude = vids
        else:                         exclude = []
        vids = [[v] for v in all_vids if v not in exclude]

    # apply simultaneously to every variable
    elif pvtype in ('all', 'all_except'):
        if pvtype.endswith('except'): exclude = vids
        else:                         exclude = []
        vids = [[v for v in all_vids if v not in exclude]]

    # apply independently to specified variables
    elif pvtype == 'independent':
        vids = [[v] for v in vids]

    # apply simultaneously to specified variables
    else:  # ptype == 'vids'
        vids = [vids]

    return procs, vids, 'independent' in pvtype


def runProcess(proc, dtable, vids, workDir, parallel):
    """Called by :func:`processData`. Runs the given process, and updates
    the :class:`.DataTable` as needed.

    :arg proc:      :class:`.Process` to run.
    :arg dtable:    :class:`.DataTable` containing the data.
    :arg workDir:   Directory to save/load new columns to/from, if the
                    processing is performed in a worker process.
    :arg vids:      List of lists, groups of variable IDs to run the process
                    on.
    :arg parallel:  If ``True``, each variable group is processed in parallel.
                    Otherwise they are processed sequentially.
    """

    results = []

    def filterMissing(vids):
        """Takes a list of variable IDs and removes those that are not
        present in the data set, emitting a warning for each ID that
        is removed.
        """
        if not proc.filterMissing:
            return vids
        filtered = []
        for vid in vids:
            if dtable.present(vid):
                filtered.append(vid)
            else:
                log.warning('Process %s refers to missing variable %u! '
                            '(%s)', proc.name, vid, proc.processString)
        return filtered

    # run process serially
    if not parallel:
        for vg in vids:
            vg = filterMissing(vg)
            results.append(proc.run(dtable, vg))

    # or run in parallel across vid groups
    else:
        with dtable.pool() as pool:

            # Note. This code is horrible for a number of
            # reasons, including that parallelisation in
            # older versions of FUNPACK worked differently,
            # and for the sake of preserving backwards
            # compatibility with resepct to the use of the
            # argument broadcast feature (although I
            # seriously doubt that anybody is even using
            # broadcasting).

            # gather all variables required by this process -
            # the ones which are specified in the processing
            # table, along with any auxillary ones speciified
            # as arguments to the process. "allvids" is used
            # to creae the sub-table, and "vids" is the list
            # of vids that we ask the processing function to
            # process.
            unfiltered = vids
            vids       = []
            allvids    = []
            bcastIdxs  = []

            for i, vg in enumerate(unfiltered):
                vg      = filterMissing(vg)
                auxvids = filterMissing(proc.auxillaryVariables(i))
                if len(vg) > 0:
                    vids     .append(vg)
                    allvids  .append(vg + auxvids)
                    bcastIdxs.append(i)

            # generate a subtable for each variable group -
            # this is to minimise the amount of data that
            # needs to be transferred to worker processes,
            # if we are parallelising.
            allcols    = [list(it.chain(*[dtable.columns(v) for v in vg]))
                          for vg in allvids]
            subtables  = [dtable.subtable(cols) for cols in allcols]

            # separate work dir for each variable group
            workDirs   = [op.join(workDir, str(i))
                          for i in range(len(vids))]
            func       = ft.partial(runParallelProcess, proc)
            parresults = pool.starmap(func, zip(subtables,
                                                vids,
                                                workDirs,
                                                bcastIdxs))

            subtables, parresults = zip(*parresults)

            results.extend(parresults)

            # Merge results back in - this
            # includes in-place modifications
            # to columns, and column flags/
            # metadata. Added/removed columns
            # are handled below.
            log.debug('Processing for %u vid groups complete - merging '
                      'results into main data table.', len(subtables))
            dtable.merge(subtables)

    remove    = []
    add       = []
    addvids   = []
    addkwargs = []

    for r in results:
        results = unpackResults(proc, r)
        remove   .extend(results[0])
        add      .extend(results[1])
        addvids  .extend(results[2])
        addkwargs.extend(results[3])

    # Parallelised processes save new
    # series to disk - load them back in.
    if log.getEffectiveLevel() >= logging.DEBUG:
        savedSeries = list(glob.glob(
            op.join(workDir, '**', '*.pkl'), recursive=True))
        if len(savedSeries) > 0:
            log.debug('[%s]: Reading %u new columns from %s %s ...',
                      proc.name, len(savedSeries), workDir, vids[:5])

    for i in range(len(add)):
        if isinstance(add[i], str):
            add[i] = pd.read_pickle(add[i])

    # remove columns first, in case
    # there is a name clash between
    # the old and new columns.
    if len(remove) > 0: dtable.removeColumns(remove)
    if len(add)    > 0: dtable.addColumns(add, addvids, addkwargs)


def runParallelProcess(proc, dtable, vids, workDir, broadcastIndex=None):
    """Used by :func:`runProcess`. Calls ``proc.run``, and returns its
    result and the (possibly modified) ``dtable``.

    :arg proc:           :class:`Process` to run
    :arg dtable:         :class:`DataTable` instance (probably a subtable)
    :arg vids:           List of variable IDs
    :arg workDir:        Directory to save new columns to if running in
                         a worker process - this is used to reduce the
                         amount of data that must be transferred between
                         processes.
    :arg broadcastIndex: Deprecated. Index to use for broadcast arguments -
                         passed through to the :meth:`Process.run` method.
    :returns:            A tuple containing:
                          - A reference to ``dtable``
                          - the result of ``proc.run()``
    """

    os.mkdir(workDir)
    result = proc.run(dtable, vids, broadcastIndex=broadcastIndex)
    remove, add, addvids, addkwargs = unpackResults(proc, result)

    # New columns are saved to disk,
    # rather than being copied back
    # to the parent process. We only
    # do this if running in a
    # multiprocessing context
    if not util.inMainProcess() and len(add) > 0:
        log.debug('[%s]: Saving %u new columns to %s %s...',
                  proc.name, len(add), workDir, vids[:5])
        for i, series in enumerate(add):
            fname  = op.join(workDir, '{}.pkl'.format(i))
            add[i] = fname
            series.to_pickle(fname)

    return dtable, (remove, add, addvids, addkwargs)


def unpackResults(proc, result):
    """Parses the results returned by a :class:`.Process`. See the
    :mod:`.processing_functions` module for details on what can be returned
    by a processing function.

    :arg proc:   The :class:`.Process`
    :arg result: The value returned by :meth:`.Process.run`.
    :returns:    A tuple containing:
                  - List of columns to remove
                  - List of new series to add
                  - List of variable IDs for new series
                  - List of :class:`.Column` keyword arguments for new series
    """

    remove    = []
    add       = []
    addvids   = []
    addkwargs = []

    if result is None:
        return remove, add, addvids, addkwargs

    error = ValueError('Invalid return value from '
                       'process {}'.format(proc.name))

    def expand(res, length):
        if res is None: return [None] * length
        else:           return res

    # columns to remove
    if isinstance(result, list):
        remove.extend(result)

    elif isinstance(result, tuple):

        # series/vids to add
        if len(result) == 2:
            add      .extend(result[0])
            addvids  .extend(expand(result[1], len(result[0])))
            addkwargs.extend(expand(None,      len(result[0])))

        # columns to remove, and
        # series/vids to add
        elif len(result) in (3, 4):

            if len(result) == 3:
                result = list(result) + [None]

            remove   .extend(result[0])
            add      .extend(result[1])
            addvids  .extend(expand(result[2], len(result[1])))
            addkwargs.extend(expand(result[3], len(result[1])))
        else:
            raise error
    else:
        raise error

    return remove, add, addvids, addkwargs


class NoSuchProcessError(Exception):
    """Exception raised by the :class:`Process` class when an unknown
    process name is specified.
    """


class Process:
    """Simple class which represents a single processing step. The :meth:`run`
    method can be used to run the process on the data for one or more
    variables.
    """


    def __init__(self, ptype, name, args, kwargs, procstr):
        """Create a ``Process``.

        :arg ptype:   Process type - either ``cleaner`` or ``processor``
                      (see the :mod:`.custom` module).
        :arg name:    Process name
        :arg args:    Positional arguments to pass to the process function.
        :arg kwargs:  Keyword arguments to pass to the process function.
        :arg procstr: Input string containing the process specification.

        Any keyword arguments which begin with ``'broadcast_'`` are separated
        out the other keyword arguments, although note that broadcasting is
        deprecated and will be removed in FUNPACK 4.0.0. See the :meth:`run`
        method for more details.
        """

        bcastKwargs  = collections.OrderedDict()
        normalKwargs = collections.OrderedDict()

        for k, v in kwargs.items():
            if k.startswith('broadcast_'):
                warnings.warn(f'[{k}]: Broadcasting is deprecated, and '
                              'will be removed in FUNPACK 4.0.0 ',
                              DeprecationWarning)
                k              = '_'.join(k.split('_')[1:])
                bcastKwargs[k] = v
            else:
                normalKwargs[k] = v

        # cleaner functions are not
        # defined in processing_functions,
        # so in this case func will be None.
        self.__ptype         = ptype
        self.__name          = name
        self.__args          = args
        self.__kwargs        = normalKwargs
        self.__procstr       = procstr
        self.__bcastKwargs   = bcastKwargs
        self.__metaproc      = normalKwargs.pop('metaproc', None)


    def __repr__(self):
        """Return a string representation of this ``Process``."""
        args    = ','.join([str(v) for v in self.__args])
        kwargs  = ','.join(['{}={}'.format(k, v) for k, v in
                           it.chain(self.__kwargs.items(),
                                    self.__bcastKwargs.items())])

        allargs = [args, kwargs]
        allargs = [a for a in allargs if a != '']
        allargs = ', '.join(allargs)
        return '{}[{}]({})'.format(self.__name, self.__ptype, allargs)


    @property
    def name(self):
        """Returns the name of this ``Process``. """
        return self.__name


    @property
    def args(self):
        """Returns the positional arguments for this ``Process``. """
        return self.__args


    @property
    def kwargs(self):
        """Returns the keyword arguments for this ``Process``. """
        return self.__kwargs


    @property
    def processString(self):
        """Returns the original string, from the processing table/
        command-line, which defines this ``Process``.
        """
        return self.__procstr


    @property
    def broadcastKwargs(self):
        """Returns the keyword arguments for this ``Process`` which
        will broadcasted across all variable IDs that are passed to
        an invocation of :meth:`run`.
        """
        return self.__bcastKwargs


    @property
    def filterMissing(self):
        """Return ``True`` if this processing function expects that the list of
        variable IDs which it is given will not contain the IDs of variables
        which are not present in the data.

        This property is set via a ``filterMissing`` argument passed to the
        processor decorator function. Its default value is ``True``.
        """
        return custom.args(self.__ptype, self.__name).get('filterMissing', True)


    def auxillaryVariables(self, broadcastIndex=None):
        """Returns a list of "auxillary" variables for this process. Auxillary
        variables are variables which a process is not being applied to, but
        which is needed by the process. These variables are passed in as
        arguments to the process.

        The names of any arguments which contain auxillary variables are
        specified via the ``auxvids`` argument to the processor decorator
        function.
        """

        auxargs = custom.args(self.__ptype, self.__name).get('auxvids', [])
        auxvids = []

        for arg in auxargs:
            if arg in self.kwargs:
                val = self.kwargs[arg]
            elif ((broadcastIndex is not None) and
                  (arg in self.broadcastKwargs)):
                val = self.broadcastKwargs[arg][broadcastIndex]
            else:
                continue

            if isinstance(val, abc.Sequence): auxvids.extend(val)
            else:                             auxvids.append(val)

        return auxvids


    def run(self, *args, broadcastIndex=None):
        """Run the process on the data, passing it the given arguments,
        and any arguments that were passed to :meth:`__init__`.

        :arg broadcastIndex: Deprecated. If provided, and if any broadcast
                             arguments were specified for this process, this
                             index is used to retrieve one value each
                             broadcast argument list - this value is then
                             passed to the process function.
        """

        kwargs = self.__kwargs.copy()
        bcast  = self.__bcastKwargs

        # retrieve the value for each broadcast argument,
        # and pass them in as regular keyword arguments
        if broadcastIndex is not None and len(bcast) > 0:
            for k, v in bcast.items():
                kwargs[k] = v[broadcastIndex]

        result = custom.run(self.__ptype,
                            self.__name,
                            *args,
                            *self.__args,
                            **kwargs)

        if self.__metaproc is not None and \
           isinstance(result, tuple)   and \
           len(result) == 4:

            # The first argument to a process
            # should be the data table
            dtable = args[0]

            # The 3rd/4th args returned from a
            # process should be a list of vids,
            # and a list of Column kwargs for
            # each of them
            vids      = result[2]
            kwargs    = result[3]
            mproc     = self.__metaproc
            newkwargs = []

            for vid, vkwargs in zip(vids, kwargs):

                if vkwargs is None or 'metadata' not in vkwargs:
                    newkwargs.append(vkwargs)
                    continue

                try:
                    vkwargs['metadata'] = custom.runMetaproc(
                        mproc, dtable, vid, vkwargs['metadata'])
                    newkwargs.append(vkwargs)

                except Exception as e:
                    log.warning('Metadata processing function '
                                'failed (vid %u): %s', vid, e)
                    newkwargs.append(vkwargs)

            result = tuple(list(result[:3]) + [newkwargs])

        return result


def parseProcesses(procs, ptype):
    """Parses the given string containing one or more comma-separated process
    calls, as defined in the processing table. Returns a list of
    :class:`Process` objects.

    :arg procs: String containing one or more comma-separated (pre-)processing
                steps.

    :arg ptype: either ``cleaner`` or ``processor``

    :returns:   A list of :class:`Process` objects.

    """

    def makeProcess(toks):
        name = toks[0]

        args   = ()
        kwargs = {}

        if len(toks) == 2:
            if isinstance(toks[1], tuple):
                args   = toks[1]
            elif isinstance(toks[1], dict):
                kwargs = toks[1]
        elif len(toks) == 3:
            args, kwargs = toks[1:]

        if not custom.exists(ptype, name):
            raise NoSuchProcessError(name)

        return Process(ptype, name, args, kwargs, procs)

    parser = pp.delimitedList(makeParser().setParseAction(makeProcess))

    try:
        parsed = parser.parseString(procs, parseAll=True)
    except Exception as e:
        log.error('Error parsing process list "{}": {}'.format(procs, e))
        raise e

    return list(parsed)


@ft.lru_cache()
def makeParser():
    """Generate a ``pyparsing`` parser which can be used to parse a single
    process call in the processing table.
    """

    lparen   = pp.Literal('(').suppress()
    rparen   = pp.Literal(')').suppress()
    lbracket = pp.Literal('[').suppress()
    rbracket = pp.Literal(']').suppress()

    def convertBoolean(tok):
        tok = tok[0]
        if   tok == 'True':  return True
        elif tok == 'False': return False
        else:                return tok

    def parseValList(toks):
        return [list(toks)]

    def parseArgs(toks):
        return [tuple(toks)]

    def parseKwargs(toks):
        kwargs = collections.OrderedDict()
        for i in range(0, len(toks), 2):
            kwargs[toks[i]] = toks[i + 1]
        return kwargs

    funcName = pp.pyparsing_common.identifier
    argval   = (pp.QuotedString('"')                                       ^
                pp.QuotedString("'")                                       ^
                pp.pyparsing_common.number                                 ^
                pp.oneOf(['True', 'False']).setParseAction(convertBoolean) ^
                pp.Literal('None').setParseAction(pp.replaceWith(None)))

    # argument values are either a
    # scalar, or a list of scalars
    # within square brackets
    vallist  = (lbracket + pp.delimitedList(argval) + rbracket)
    vallist  = vallist.setParseAction(parseValList)
    argval   = argval ^ vallist

    # arguments are either
    # positional or keyword
    kwargs   = (pp.pyparsing_common.identifier +
                pp.Literal('=').suppress() +
                argval)
    posargs  = pp.delimitedList(argval).setParseAction(parseArgs)
    kwargs   = pp.delimitedList(kwargs).setParseAction(parseKwargs)
    allargs  = pp.delimitedList(pp.Optional(posargs) + pp.Optional(kwargs))

    # function can be called as:
    #   function
    #   function()
    #   function(args)
    allargs  = lparen   + pp.Optional(allargs) + rparen
    function = funcName + pp.Optional(allargs)

    return function



#!/usr/bin/env python
#
# test_metaproc.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import random
import textwrap as tw
from unittest import mock

import numpy as np

from .. import metaproc_functions as metaproc
from .. import hierarchy
from .. import custom

from . import clear_plugins, gen_DataTable, tempdir, patch_logging


@clear_plugins
@patch_logging
def test_codingDescriptionFromValue():

    custom.registerBuiltIns()

    with tempdir():
        descs = tw.dedent("""
        coding	meaning
        30	meaning 30
        40	meaning 40
        50	meaning 50
        20	meaning 20
        60	meaning 60
        10	meaning 10 hah
        """).strip()

        with open('coding123.txt', 'wt') as f:
            f.write(descs)

        dt = gen_DataTable([np.random.randint(1, 10, 10)])

        dt.vartable.loc[1, 'DataCoding'] = 123

        func = custom.get('metaproc', 'codingdesc')

        with mock.patch('funpack.coding.getCodingFilePath',
                        return_value='coding123.txt'):
            assert func(dt, 1, 30) == '30 - meaning 30'
            assert func(dt, 1, 10) == '10 - meaning 10 hah'


@clear_plugins
@patch_logging
def test_hierarchicalDescriptionFromX():

    custom.registerBuiltIns()

    with tempdir():
        data = tw.dedent("""
        coding\tmeaning\tnode_id\tparent_id
        a\ta desc\t5\t0
        b\tb desc\t1\t5
        c\tc desc\t3\t5
        d\td desc\t4\t3
        e\te desc\t2\t1
        """)

        with open('coding19.txt', 'wt') as f:
            f.write(data)

        dt = gen_DataTable([np.random.randint(1, 10, 10)])

        dt.vartable.loc[1, 'DataCoding'] = 123

        numfunc  = custom.get('metaproc', 'hierarchynumdesc')
        codefunc = custom.get('metaproc', 'hierarchycodedesc')

        with mock.patch('funpack.hierarchy.getHierarchyFilePath',
                        return_value='coding19.txt'):
            assert codefunc(dt, 1, 'a') == 'a - a desc'
            assert codefunc(dt, 1, 'c') == 'c - c desc'
            assert numfunc( dt, 1,  5)  == 'a - a desc'
            assert numfunc( dt, 1,  3)  == 'c - c desc'



#!/usr/bin/env python
#
# exporting.py - Functions for exporting data
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions for exporting data to a file. """


import                     logging
import                     collections

import numpy            as np
import pandas.api.types as pdtypes

from . import util
from . import custom


log = logging.getLogger(__name__)


def exportData(dtable, outfile, fileFormat, **kwargs):
    """Export the data contained in ``dtable`` to ``outfile`` using the
    specified ``fileFormat``.

    :arg dtable:        :class:`.DataTable` containing the data to export.

    :arg outfile:       File to export data to.

    :arg fileFormat:    File format to export to - the name of a ``@exporter``
                        plugin.

    All other arguments are passed through to the exporter plugin.
    """

    if dtable.shape[0] == 0 or dtable.shape[1] == 0:
        raise RuntimeError('No data to export (rows: {}, columns: '
                           '{})'.format(*dtable.shape))

    custom.runExporter(fileFormat, dtable, outfile, **kwargs)


def formatColumn(col,
                 dtable,
                 dateFormat,
                 timeFormat,
                 formatters,
                 logmsg=None):
    """Formats the data for the specified column. The ``dtable`` is updated
    in-place.

    :arg col:        :class:`.Column` to format

    :arg dtable:     :class:`.DataTable` containing the data

    :arg dateFormat: Name of formatter to use for date columns.

    :arg timeFormat: Name of formatter to use for time columns.

    :arg formatters: Dict of ``{ [vid|column] : formatter }`` mappings,
                     specifying custom formatters to use for specific
                     variables.

    :arg logmsg:     Message to include in log message.

    :returns:        A reference to ``dtable``.
    """

    if dateFormat is None: dateFormat = 'default'
    if timeFormat is None: timeFormat = 'default'
    if formatters is None: formatters = {}

    if logmsg is None: logmsg = ''
    else:              logmsg = ' [{}]'.format(logmsg)

    vid      = col.basevid
    vartable = dtable.vartable
    series   = dtable[:, col.name]

    # formatters can be specified
    # by VID or by column name
    formatter = formatters.get(vid, None)
    if formatter is None:
        formatter = formatters.get(col.name, None)

    if vid in vartable.index: vtype = vartable['Type'][vid]
    else:                     vtype = None

    # fall back to date/time formatting
    # if relevant for this column
    if formatter is None and pdtypes.is_datetime64_any_dtype(series):
        # use dateFormat if we know the column
        # is a date (and not datetime), otherwise
        # use timeFormat if the column is a
        # datetime, or unknown type.
        if vtype == util.CTYPES.date: formatter = dateFormat
        else:                         formatter = timeFormat

    if formatter is not None:
        log.debug('Formatting column %s%s with %s formatter',
                  col.name, logmsg, formatter)
        series = custom.runFormatter(formatter, dtable, col, series)

    # apply column-specific fill
    # value, if there is one
    if col.fillval is not None:
        series = series.fillna(value=col.fillval)

    # update datatable if any
    # formatting took place
    if (formatter is not None) or (col.fillval is not None):
        dtable[:, col.name] = series

    return dtable


@custom.formatter('default')
def defaultDateTimeFormat(dtable, column, series):
    """Default format converter for date and time columns. """

    if pdtypes.is_datetime64_any_dtype(series):

        # pandas uses the same data type
        # (pandas.Timestamp) for all date/time
        # types. So to distinguish between
        # dates, and full time stamps, we
        # need to look at the biobank type
        vttype = dtable.vartable.loc[column.basevid, 'Type']

        # The biobank "Date" type is just a date
        if vttype == util.CTYPES.date: fmtstr = '%Y-%m-%d'
        else:                          fmtstr = '%Y-%m-%dT%H:%M:%S%z'

        def fmt(val):
            try:
                return val.strftime(fmtstr)
            except Exception:
                return np.nan

        return series.apply(fmt)

    else:
        return series


@custom.formatter('compound')
def formatCompound(dtable, column, series, delim=','):
    """Format a compound (multi-valued) column which is stored in-memory
    as a list or ``numpy`` array.
    """

    if len(series) == 0:
        return series

    sample = series.iloc[0]

    if isinstance(sample, str):
        return series
    if not isinstance(sample, (np.ndarray, collections.abc.Sequence)):
        return series

    def fmt(val):
        return delim.join([str(v) for v in val])

    return series.apply(fmt)


@custom.formatter('escapeString')
def escapeString(dtable, column, series):
    """Encodes every value of ``series`` as a string, with any escape
    characters (``\\n``, ``\\t``, etc) formatted literally. This has the
    effect that the series is coerced to a string type (if not already).
    """
    if len(series) == 0:
        return series

    def fmt(val):
        return str(val).encode('unicode_escape').decode('utf-8')

    return series.apply(fmt)



#!/usr/bin/env python

import shutil
import sys
import os.path as op

from setuptools import setup, find_namespace_packages, Command

basedir = op.dirname(__file__)


with open(op.join(basedir, 'requirements.txt'), 'rt') as f:
    install_requires = [line.strip() for line in f.readlines()]
    install_requires = [line for line in install_requires if line != '']


with open(op.join(basedir, 'requirements-demo.txt'), 'rt') as f:
    demo_requires = [line.strip() for line in f.readlines()]
    demo_requires = [line for line in demo_requires if line != '']


with open(op.join(basedir, 'requirements-test.txt'), 'rt') as f:
    test_requires = [line.strip() for line in f.readlines()]
    test_requires = [line for line in demo_requires if line != '']


extras_require = {
    'demo' : demo_requires,
    'test' : test_requires
}


version = {}
with open(op.join(basedir, 'funpack', '__init__.py')) as f:
    for line in f:
        if line.startswith('__version__ = '):
            exec(line, version)
            break
version = version['__version__']


with open(op.join(basedir, 'README.rst'), 'rt') as f:
    readme = f.read()


class doc(Command):
    """Build the API documentation. """

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):

        docdir  = op.join(op.dirname(__file__), 'doc')
        destdir = op.join(docdir, 'html')

        # so sphinx can find our custom extensions
        sys.path.append(docdir)

        if op.exists(destdir):
            shutil.rmtree(destdir)

        import funpack.scripts.generate_notebooks as gennb
        import funpack.config                     as config
        import funpack.custom                     as custom
        import funpack.util                       as util
        import sphinx.cmd.build                   as sphinx_build

        # generate doc/demo.ipynb
        demofile = op.join(basedir, 'doc', 'demo.ipynb')
        if not op.exists(demofile):
            gennb.main()
        sphinx_build.main([docdir, destdir])

setup(
    name='fmrib-unpack',
    version=version,
    description='The FMRIB UKBiobank Normalisation, Parsing '
                'And Cleaning Kit',
    long_description=readme,
    url='https://git.fmrib.ox.ac.uk/fsl/funpack',
    author='Paul McCarthy',
    author_email='pauldmccarthy@gmail.com',
    license='Apache License Version 2.0',

    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],

    install_requires=install_requires,
    extras_require=extras_require,
    packages=find_namespace_packages(include=('funpack', 'funpack.*')),
    include_package_data=True,

    cmdclass={'doc' : doc},

    entry_points={
        'console_scripts' : [
            'fmrib_unpack      = funpack.main:main',
            'fmrib_unpack_demo = funpack.scripts.demo:main',
        ]
    }
)



#!/usr/bin/env python
#
# test_main_auxillary_output_files.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import textwrap  as tw
import itertools as it
import os.path   as op
import              os
import              random

from unittest import mock

import numpy as np

import pandas as pd

import funpack.main       as main
import funpack.util       as util
import funpack.icd10      as icd10
import funpack.custom     as custom
import funpack.expression as expression
import funpack.processing as processing

from . import (patch_logging,
               tempdir,
               gen_DataTable,
               gen_test_data)


@patch_logging
def test_main_icd10():
    with tempdir():
        codings = tw.dedent("""
        coding\tmeaning\tnode_id\tparent_id
        a10\ta desc\t5\t0
        b20\tb desc\t1\t5
        c30\tc desc\t3\t5
        d40\td desc\t4\t3
        e50\te desc\t2\t1
        """).strip()

        data = tw.dedent("""
        eid,1-0.0
        1,a10
        2,b20
        3,c30
        4,d40
        5,e50
        """)

        exp = tw.dedent("""
        code\tvalue\tdescription\tparent_descs
        a10\t5\ta desc\t
        b20\t1\tb desc\t[a desc]
        c30\t3\tc desc\t[a desc]
        d40\t4\td desc\t[a desc] [c desc]
        e50\t2\te desc\t[a desc] [b desc]
        """).strip()

        with open('coding19.tsv', 'wt') as f: f.write(codings)
        with open('data.tsv',     'wt') as f: f.write(data)

        with mock.patch('funpack.hierarchy.getHierarchyFilePath',
                        return_value='coding19.tsv'), \
            mock.patch.dict('funpack.util.DATA_TYPES',
                             {util.CTYPES.unknown : str}):
            main.main('-nb -cl 1 codeToNumeric(\'icd10\') '
                      '-imf icd10_mappings.tsv out.tsv data.tsv'
                      .split())

        with open('icd10_mappings.tsv', 'rt') as f:
            got = f.read().strip()

        assert exp == got
    icd10.storeCodes.store = []


@patch_logging
def test_main_unknown_vars():

    vartable = tw.dedent("""
    ID\tType\tDescription\tDataCoding\tNAValues\tRawLevels\tNewLevels\tParentValues\tChildValues\tClean
    1
    2
    3
    4
    5
    6
    7
    8
    9
    10
    """).strip()

    cattable = tw.dedent("""
    ID\tCategory\tVariables
    1\tknown\t1:5
    """).strip()

    exp = tw.dedent("""
    name\tfile\tclass\texported
    11-0.0\t{file}\tunknown\t{}
    12-0.0\t{file}\tunknown\t{}
    13-0.0\t{file}\tunknown\t{}
    14-0.0\t{file}\tunknown\t{}
    15-0.0\t{file}\tunknown\t{}
    6-0.0\t{file}\tuncategorised\t{}
    7-0.0\t{file}\tuncategorised\t{}
    8-0.0\t{file}\tuncategorised\t{}
    9-0.0\t{file}\tuncategorised\t{}
    10-0.0\t{file}\tuncategorised\t{}
    """).strip()

    def check(fname, *fmtargs, **fmtkwargs):

        drop = fmtkwargs.pop('drop', [])

        cexp = exp.split('\n')

        for d in drop:
            cexp = [l for l in cexp if not l.startswith('{}-'.format(d))]
        cexp = '\n'.join(cexp)

        with open(fname, 'rt') as f:
            got = f.read().strip()

        print('exp', cexp.format(*fmtargs, **fmtkwargs))
        print('got')
        print( got)

        assert got == cexp.format(*fmtargs, **fmtkwargs)

    with tempdir() as td:

        fullfile = op.realpath(op.join(td, 'data.tsv'))

        gen_test_data(15, 50, 'data.tsv')
        with open('variables.tsv', 'wt') as f:
            f.write(vartable)
        with open('categories.tsv', 'wt') as f:
            f.write(cattable)

        # only import known vars -
        # no file generated
        main.main('-ow -nb -wu -c known '
                  '-vf variables.tsv '
                  '-cf categories.tsv '
                  'out.tsv data.tsv'.split())
        assert not op.exists('out_unknown_vars.txt')

        # import all vars - we get
        # a file with all unknowns/uncategorised
        main.main('-ow -nb -wu '
                  '-vf variables.tsv '
                  '-cf categories.tsv '
                  'out.tsv data.tsv'.split())
        check('out_unknown_vars.txt',
              1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
              file=fullfile)
        os.remove('out_unknown_vars.txt')

        # repeat above, specifying custom filename
        main.main('-ow -nb  '
                  '-vf variables.tsv '
                  '-uf unknowns.tsv '
                  '-cf categories.tsv '
                  'out.tsv data.tsv'.split())
        check('unknowns.tsv',
              1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
              file=fullfile)
        os.remove('unknowns.tsv')

        # some unknowns/uncats imported
        main.main('-ow -nb -wu '
                  '-vf variables.tsv '
                  '-v 1:3 -v 6:8 -v 11:13 '
                  '-cf categories.tsv '
                  'out.tsv data.tsv'.split())
        check('out_unknown_vars.txt',
              1, 1, 1, 1, 1, 1,
              file=fullfile, drop=[9, 10, 14, 15])
        os.remove('out_unknown_vars.txt')

        # some unknowns/uncats
        # failed processing
        data2 = pd.read_csv('data.tsv', delimiter='\t', index_col=0)
        data2.loc[1:45, '6-0.0']  = np.nan
        data2.loc[5:,   '11-0.0'] = np.nan
        data2.to_csv('data2.tsv', sep='\t')
        fullfile2 = op.realpath(op.join(td, 'data2.tsv'))
        main.main('-ow -nb -wu '
                  '-apr 6  removeIfSparse(minpres=20) '
                  '-apr 11 removeIfSparse(minpres=20) '
                  '-vf variables.tsv '
                  '-cf categories.tsv '
                  'out.tsv data2.tsv'.split())
        check('out_unknown_vars.txt',
              0, 1, 1, 1, 1, 0, 1, 1, 1, 1,
              file=fullfile2)
        os.remove('out_unknown_vars.txt')

        # unknown/uncat vars with
        # processing defined - should not
        # affect classification
        main.main('-ow -nb -wu '
                  '-nv 6 1,2,3 '
                  '-nv 11 1,2,3 '
                  '-vf variables.tsv '
                  '-cf categories.tsv '
                  'out.tsv data.tsv'.split())
        check('out_unknown_vars.txt',
              1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
              file=fullfile)
        os.remove('out_unknown_vars.txt')



@patch_logging
def test_main_description_file():

    vartable = tw.dedent("""
    ID\tType\tDescription\tDataCoding\tNAValues\tRawLevels\tNewLevels\tParentValues\tChildValues\tClean
    1\t\tvar one
    2\t\tvar two
    3\t\tvar three
    4
    """).strip()

    proctable = tw.dedent("""
    Variable\tProcess
    1\tmyprocess
    2\tmyprocess(metaproc='mymetaproc')
    """)

    @custom.processor()
    def myprocess(dtable, vids):
        columns = it.chain(*[dtable.columns(v) for v in vids])

        add     = []
        addvid  = []
        addmeta = []

        for col in columns:
            series = dtable[:, col.name]
            col.metadata = col.vid + 20

            newseries = pd.Series(series + 5, name=col.name + '_a')
            dtable[:, col.name] = dtable[:, col.name] + 3

            add    .append(newseries)
            addvid .append(col.vid)
            addmeta.append({'metadata' : col.vid + 40})

        return [], add, addvid, addmeta

    @custom.metaproc()
    def mymetaproc(dtable, vid, val):
        return str(val) + ' metaprocced'

    with tempdir():
        with open('vartable.tsv',  'wt') as f: f.write(vartable)
        with open('proctable.tsv', 'wt') as f: f.write(proctable)

        gen_test_data(4, 10, 'data.tsv')

        main.main('-nb -vf vartable.tsv -pf proctable.tsv '
                  '-def descriptions.tsv '
                  'out.tsv data.tsv'.split())

        inp   = pd.read_csv('data.tsv',         delimiter='\t', index_col=0)
        got   = pd.read_csv('out.tsv',          delimiter='\t', index_col=0)
        descs = pd.read_csv('descriptions.tsv',
                            delimiter='\t',
                            index_col=0,
                            header=None,
                            names=('column', 'description'))

        assert sorted(got.columns) == ['1-0.0', '1-0.0_a',
                                       '2-0.0', '2-0.0_a',
                                       '3-0.0', '4-0.0']

        assert (got['1-0.0']   == (inp['1-0.0'] + 3)).all()
        assert (got['1-0.0_a'] == (inp['1-0.0'] + 5)).all()
        assert (got['2-0.0']   == (inp['2-0.0'] + 3)).all()
        assert (got['2-0.0_a'] == (inp['2-0.0'] + 5)).all()
        assert (got['3-0.0']   == (inp['3-0.0']))    .all()
        assert (got['4-0.0']   == (inp['4-0.0']))    .all()

        assert descs.loc['1-0.0',   'description'] == 'var one (21)'
        assert descs.loc['1-0.0_a', 'description'] == 'var one (41)'
        assert descs.loc['2-0.0',   'description'] == 'var two (22)'
        assert descs.loc['2-0.0_a', 'description'] == 'var two (42 metaprocced)'
        assert descs.loc['3-0.0',   'description'] == 'var three (0.0)'
        assert descs.loc['4-0.0',   'description'] == 'n/a (0.0)'


@patch_logging
def test_main_summary_file():
    class A:
        pass
    with tempdir():
        data = np.random.randint(1, 10, (10, 4))
        dt   = gen_DataTable(data)
        vt   = dt.vartable

        args = A()
        args.summary_file = 'summary.tsv'

        vt.at[1, 'RawLevels']    = [1, 2, 3]
        vt.at[1, 'NewLevels']    = [3, 2, 1]
        vt.at[2, 'NAValues']     = [1, 2, 3, 4]
        vt.at[3, 'ParentValues'] = [expression.Expression('v1 == 0')]
        vt.at[3, 'ChildValues']  = [0]
        vt.at[4, 'Clean']        = {
            'flattenHierarchical' :
            processing.Process('cleaner', 'flattenHierarchical', (), {},
                               'procstr')}

        main.doSummaryExport(dt, args)

        sum = pd.read_csv('summary.tsv', delimiter='\t', index_col=0)

        assert sum.at[1, 'RawLevels']    == str(vt.at[1, 'RawLevels'])
        assert sum.at[1, 'NewLevels']    == str(vt.at[1, 'NewLevels'])
        assert sum.at[2, 'NAValues']     == str(vt.at[2, 'NAValues'])
        assert sum.at[3, 'ParentValues'] == str(vt.at[3, 'ParentValues'])
        assert sum.at[3, 'ChildValues']  == str(vt.at[3, 'ChildValues'])
        assert sum.at[4, 'Clean']        == '[{}]'.format(
            str(vt.at[4, 'Clean']['flattenHierarchical']))


@patch_logging
def test_main_default_auxfile_names():

    with tempdir():

        gen_test_data(20, 20, 'data.tsv')
        main.main('-wl -wu -wim -wde -ws '
                  'my_output.tsv data.tsv'.split())

        assert op.exists('my_output.tsv')
        assert op.exists('my_output_log.txt')
        assert op.exists('my_output_unknown_vars.txt')
        assert op.exists('my_output_icd10_map.txt')
        assert op.exists('my_output_description.txt')
        assert op.exists('my_output_summary.txt')

    with tempdir():

        gen_test_data(20, 20, 'data.tsv')
        main.main('-wl -wu -wim -wde -ws '
                  '-lf mylog.txt '
                  '-uf myunknowns.txt '
                  '-imf myicd10s.txt '
                  '-def mydesc.txt '
                  '-sf mysum.txt '
                  'my_output.tsv data.tsv'.split())

        assert op.exists('my_output.tsv')
        assert op.exists('mylog.txt')
        assert op.exists('myunknowns.txt')
        assert op.exists('myicd10s.txt')
        assert op.exists('mydesc.txt')
        assert op.exists('mysum.txt')
